"""
Test utilities for :py:mod:`betty.ancestry.gender`.
"""

from __future__ import annotations

from betty.ancestry.gender import Gender
from betty.test_utils.plugin import (
    DummyPlugin,
    PluginTestBase,
)


class GenderTestBase(PluginTestBase[Gender]):
    """
    A base class for testing :py:class:`betty.ancestry.gender.Gender` implementations.
    """

    pass


class DummyGender(DummyPlugin, Gender):
    """
    A dummy gender implementation.
    """

    pass



"""
Data types that have media types.
"""

from __future__ import annotations

from typing import Any, TYPE_CHECKING

from typing_extensions import override

from betty.json.linked_data import LinkedDataDumpableJsonLdObject, JsonLdObject
from betty.media_type import MediaType, MediaTypeSchema
from betty.privacy import is_public

if TYPE_CHECKING:
    from betty.serde.dump import DumpMapping, Dump
    from betty.project import Project


class HasMediaType(LinkedDataDumpableJsonLdObject):
    """
    A resource with an `IANA media type <https://www.iana.org/assignments/media-types/media-types.xhtml>`_.
    """

    def __init__(
        self,
        *args: Any,
        media_type: MediaType | None = None,
        **kwargs: Any,
    ):
        super().__init__(*args, **kwargs)
        self.media_type = media_type

    @override
    async def dump_linked_data(self, project: Project) -> DumpMapping[Dump]:
        dump = await super().dump_linked_data(project)
        if is_public(self) and self.media_type is not None:
            dump["mediaType"] = str(self.media_type)
        return dump

    @override
    @classmethod
    async def linked_data_schema(cls, project: Project) -> JsonLdObject:
        schema = await super().linked_data_schema(project)
        schema.add_property("mediaType", MediaTypeSchema(), False)
        return schema



"""
Test utilities for :py:mod:`betty.cache`.
"""

import asyncio
import multiprocessing
import pickle
from asyncio import sleep, create_task
from typing import Generic, Sequence, AsyncContextManager, Iterator, TypeVar, Any

import pytest

from betty.cache import Cache

_CacheItemValueT = TypeVar("_CacheItemValueT")


class CacheTestBase(Generic[_CacheItemValueT]):
    """
    A base class for tests of :py:class:`betty.cache.Cache` implementations.
    """

    def _new_sut(
        self,
        *,
        scopes: Sequence[str] | None = None,
    ) -> AsyncContextManager[Cache[_CacheItemValueT]]:
        raise NotImplementedError

    def _values(self) -> Iterator[_CacheItemValueT]:
        raise NotImplementedError

    async def test_with_scope(self) -> None:
        """
        Test implementations of :py:meth:`betty.cache.Cache.with_scope`.
        """
        for value in self._values():
            async with self._new_sut() as sut:
                sut_with_scope_one = sut.with_scope("scopey")
                sut_with_scope_two = sut.with_scope("scopey")
                assert sut_with_scope_one is not sut
                assert sut_with_scope_two is not sut
                cache_item_id = "hello-world"
                await sut_with_scope_one.set(cache_item_id, value)
                async with sut_with_scope_two.get(cache_item_id) as cache_item:
                    assert cache_item
                    assert await cache_item.value() == value

    @pytest.mark.parametrize(
        "scopes",
        [
            (),
            ("scopey", "dopey"),
        ],
    )
    async def test_get_without_hit(self, scopes: Sequence[str]) -> None:
        """
        Test implementations of :py:meth:`betty.cache.Cache.get`.
        """
        async with self._new_sut(scopes=scopes) as sut, sut.get("id") as cache_item:
            assert cache_item is None

    @pytest.mark.parametrize(
        "scopes",
        [
            (),
            ("scopey", "dopey"),
        ],
    )
    async def test_set_and_get(self, scopes: Sequence[str]) -> None:
        """
        Test implementations of :py:meth:`betty.cache.Cache.get` and :py:meth:`betty.cache.Cache.set`.
        """
        for value in self._values():
            async with self._new_sut(scopes=scopes) as sut:
                await sut.set("id", value)
                async with sut.get("id") as cache_item:
                    assert cache_item is not None
                    assert await cache_item.value() == value

    @pytest.mark.parametrize(
        "scopes",
        [
            (),
            ("scopey", "dopey"),
        ],
    )
    async def test_set_and_get_with_modified(self, scopes: Sequence[str]) -> None:
        """
        Test implementations of :py:meth:`betty.cache.Cache.get` and :py:meth:`betty.cache.Cache.set`.
        """
        modified = 123456789
        for value in self._values():
            async with self._new_sut(scopes=scopes) as sut:
                await sut.set("id", value, modified=modified)
                async with sut.get("id") as cache_item:
                    assert cache_item is not None
                    assert cache_item.modified == modified

    @pytest.mark.parametrize(
        "scopes",
        [
            (),
            ("scopey", "dopey"),
        ],
    )
    async def test_getset_without_hit(self, scopes: Sequence[str]) -> None:
        """
        Test implementations of :py:meth:`betty.cache.Cache.getset`.
        """
        for value in self._values():
            async with self._new_sut(scopes=scopes) as sut:
                async with sut.getset("id") as (cache_item, setter):
                    assert cache_item is None
                    await setter(value)
                async with sut.get("id") as cache_item:
                    assert cache_item is not None
                    assert await cache_item.value() == value

    @pytest.mark.parametrize(
        "scopes",
        [
            (),
            ("scopey", "dopey"),
        ],
    )
    async def test_getset_with_hit(self, scopes: Sequence[str]) -> None:
        """
        Test implementations of :py:meth:`betty.cache.Cache.getset`.
        """
        for value in self._values():
            async with self._new_sut(scopes=scopes) as sut:
                await sut.set("id", value)
                async with sut.getset("id") as (cache_item, setter):
                    assert cache_item is not None
                    assert await cache_item.value() == value

    @pytest.mark.parametrize(
        "scopes",
        [
            (),
            ("scopey", "dopey"),
        ],
    )
    async def test_getset_without_lock(self, scopes: Sequence[str]) -> None:
        """
        Test implementations of :py:meth:`betty.cache.Cache.getset`.
        """
        async with self._new_sut(scopes=scopes) as sut:

            async def _acquire() -> None:
                async with sut.getset("id"):
                    await sleep(999)

            task = create_task(_acquire())
            await sleep(1)
            try:
                async with sut.getset("id", wait=False) as (cache_item, setter):
                    assert cache_item is None
                    assert setter is None
            finally:
                task.cancel()

    @pytest.mark.parametrize(
        "scopes",
        [
            (),
            ("scopey", "dopey"),
        ],
    )
    async def test_delete(self, scopes: Sequence[str]) -> None:
        """
        Test implementations of :py:meth:`betty.cache.Cache.delete`.
        """
        async with self._new_sut(scopes=scopes) as sut:
            await sut.set("id", next(self._values()))
            await sut.delete("id")
            async with sut.get("id") as cache_item:
                assert cache_item is None

    @pytest.mark.parametrize(
        "scopes",
        [
            (),
            ("scopey", "dopey"),
        ],
    )
    async def test_clear(self, scopes: Sequence[str]) -> None:
        """
        Test implementations of :py:meth:`betty.cache.Cache.clear`.
        """
        async with self._new_sut(scopes=scopes) as sut:
            await sut.set("id", next(self._values()))
            await sut.clear()
            async with sut.get("id") as cache_item:
                assert cache_item is None


class ProcesssafeCacheTestBase(
    Generic[_CacheItemValueT], CacheTestBase[_CacheItemValueT]
):
    """
    A base class for tests of :py:class:`betty.cache.Cache` implementations that are process-safe.
    """

    async def test_pickle(self) -> None:
        """
        Test that implementations can be pickled.
        """
        async with self._new_sut() as sut:
            pickle.loads(pickle.dumps(sut))

    @classmethod
    def _test_get_with_multiprocessing_target(
        cls, sut: Cache[_CacheItemValueT], cache_item_id: str, value: _CacheItemValueT
    ) -> None:
        asyncio.run(sut.set(cache_item_id, value))

    async def test_set_with_multiprocessing(self) -> None:
        """
        Test implementations of :py:meth:`betty.cache.Cache.set`.
        """
        for value in self._values():
            async with self._new_sut() as sut:
                process = multiprocessing.Process(
                    target=type(self)._test_get_with_multiprocessing_target,
                    args=(sut, "id", value),
                )
                process.start()
                process.join()
                async with sut.get("id") as cache_item:
                    assert cache_item
                    assert await cache_item.value() == value

    @classmethod
    def _test_delete_with_multiprocessing_target(
        cls, sut: Cache[_CacheItemValueT], cache_item_id: str
    ) -> None:
        asyncio.run(sut.delete(cache_item_id))

    async def test_delete_with_multiprocessing(self) -> None:
        """
        Test implementations of :py:meth:`betty.cache.Cache.delete`.
        """
        async with self._new_sut() as sut:
            await sut.set("id", next(self._values()))
            process = multiprocessing.Process(
                target=type(self)._test_delete_with_multiprocessing_target,
                args=(sut, "id"),
            )
            process.start()
            process.join()
            async with sut.get("id") as cache_item:
                assert cache_item is None

    @classmethod
    def _test_clear_with_multiprocessing_target(cls, sut: Cache[Any]):
        asyncio.run(sut.clear())

    async def test_clear_with_multiprocessing(self) -> None:
        """
        Test implementations of :py:meth:`betty.cache.Cache.clear`.
        """
        async with self._new_sut() as sut:
            await sut.set("id", next(self._values()))
            process = multiprocessing.Process(
                target=type(self)._test_clear_with_multiprocessing_target, args=(sut,)
            )
            process.start()
            process.join()
            async with sut.get("id") as cache_item:
                assert cache_item is None



"""
Provide the Render API.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import final, TYPE_CHECKING

from typing_extensions import override

from betty.plugin.entry_point import EntryPointPluginRepository
from betty.typing import internal

if TYPE_CHECKING:
    from betty.plugin import PluginRepository, Plugin
    from betty.locale.localizer import Localizer
    from betty.job import Context
    from pathlib import Path
    from collections.abc import Sequence


class Renderer(ABC):
    """
    Render content to HTML.

    Read more about :doc:`/development/plugin/renderer`.
    """

    @property
    @abstractmethod
    def file_extensions(self) -> set[str]:
        """
        The extensions of the files this renderer can render.
        """
        pass

    @abstractmethod
    async def render_file(
        self,
        file_path: Path,
        *,
        job_context: Context | None = None,
        localizer: Localizer | None = None,
    ) -> Path:
        """
        Render a single file.

        :return: The file's new path, which may have been changed, e.g. a
            renderer-specific extension may have been stripped from the end.
        """
        pass


RENDERER_REPOSITORY: PluginRepository[Renderer & Plugin] = EntryPointPluginRepository(
    "betty.renderer"
)
"""
The renderer plugin repository.

Read more about :doc:`/development/plugin/renderer`.
"""


@internal
@final
class SequentialRenderer(Renderer):
    """
    Render using a sequence of other renderers.
    """

    def __init__(self, renderers: Sequence[Renderer]):
        self._renderers = renderers
        self._file_extensions = {
            file_extension
            for renderer in self._renderers
            for file_extension in renderer.file_extensions
        }

    @override
    @property
    def file_extensions(self) -> set[str]:
        return self._file_extensions

    @override
    async def render_file(
        self,
        file_path: Path,
        *,
        job_context: Context | None = None,
        localizer: Localizer | None = None,
    ) -> Path:
        for renderer in self._renderers:
            for renderer_file_extension in renderer.file_extensions:
                if file_path.suffix.endswith(renderer_file_extension):
                    return await renderer.render_file(
                        file_path,
                        job_context=job_context,
                        localizer=localizer,
                    )
        return file_path



"""
Provide static plugin management.
"""

from collections.abc import AsyncIterator
from typing import Generic, TypeVar, final

from typing_extensions import override

from betty.factory import Factory
from betty.machine_name import MachineName
from betty.plugin import PluginRepository, Plugin, PluginNotFound

_PluginT = TypeVar("_PluginT", bound=Plugin)


@final
class StaticPluginRepository(PluginRepository[_PluginT], Generic[_PluginT]):
    """
    A repository that is given a static collection of plugins, and exposes those.
    """

    def __init__(self, *plugins: type[_PluginT], factory: Factory | None = None):
        super().__init__(factory=factory)
        self._plugins = {plugin.plugin_id(): plugin for plugin in plugins}

    @override
    async def get(self, plugin_id: MachineName) -> type[_PluginT]:
        try:
            return self._plugins[plugin_id]
        except KeyError:
            raise PluginNotFound.new(plugin_id, await self.select()) from None

    @override
    async def __aiter__(self) -> AsyncIterator[type[_PluginT]]:
        for plugin in self._plugins.values():
            yield plugin



"""
Test utilities for :py:mod:`betty.model.collections`.
"""

from collections.abc import Sequence
from typing import TypeVar, Generic, Any

import pytest

from betty.model import Entity
from betty.model.collections import EntityCollection

_EntityT = TypeVar("_EntityT", bound=Entity)
_EntityCollectionT = TypeVar("_EntityCollectionT", bound=EntityCollection[Entity])


class EntityCollectionTestBase(Generic[_EntityT]):
    """
    A base class for testing :py:class:`betty.model.collections.EntityCollection` implementations.
    """

    async def get_suts(self) -> Sequence[EntityCollection[_EntityT]]:
        """
        Produce the (empty) collections under test.

        This MUST return at least one entity collection.
        """
        raise NotImplementedError

    async def get_entities(self) -> Sequence[_EntityT]:
        """
        Produce entities to test the collections with.

        This MUST return at least 3 entities.
        """
        raise NotImplementedError

    async def test_entity_collection_test_base_get_entities(self) -> None:
        """
        Tests :py:meth:`betty.test_utils.model.collections.EntityCollectionTestBase.get_entities` implementations.
        """
        assert len(await self.get_entities()) >= 3

    async def test_add(self) -> None:
        """
        Tests :py:meth:`betty.model.collections.EntityCollection.add` implementations.
        """
        for sut in await self.get_suts():
            entities = await self.get_entities()
            sut.add(*entities)
            assert list(sut) == list(entities)

    async def test_add_with_duplicate_entities(self) -> None:
        """
        Tests :py:meth:`betty.model.collections.EntityCollection.add` implementations.
        """
        for sut in await self.get_suts():
            entities = await self.get_entities()
            sut.add(entities[0], entities[1], entities[0], entities[2])
            assert list(sut) == list(entities[0:3])

    async def test_remove(self) -> None:
        """
        Tests :py:meth:`betty.model.collections.EntityCollection.remove` implementations.
        """
        for sut in await self.get_suts():
            entities = await self.get_entities()
            sut.add(*entities)
            first = entities[0]
            sut.add(*entities)
            sut.remove(first)
            assert list(sut) == list(entities[1:])
            sut.remove(*entities)
            assert list(sut) == []

    async def test___delitem____by_entity(self) -> None:
        """
        Tests :py:meth:`betty.model.collections.EntityCollection.__delitem__` implementations.
        """
        for sut in await self.get_suts():
            entities = await self.get_entities()
            sut.add(*entities)
            first = entities[0]
            sut.add(*entities)
            del sut[first]
            assert list(sut) == list(entities[1:])

    async def test___contains____by_entity(self) -> None:
        """
        Tests :py:meth:`betty.model.collections.EntityCollection.__contains__` implementations.
        """
        for sut in await self.get_suts():
            entities = await self.get_entities()
            sut.add(entities[0])
            assert entities[0] in sut
            assert entities[1] not in sut

    @pytest.mark.parametrize(
        "value",
        [
            True,
            False,
            [],
        ],
    )
    async def test___contains____by_unsupported_type(self, value: Any) -> None:
        """
        Tests :py:meth:`betty.model.collections.EntityCollection.__contains__` implementations.
        """
        for sut in await self.get_suts():
            assert value not in sut

    async def test___len__(self) -> None:
        """
        Tests :py:meth:`betty.model.collections.EntityCollection.__len__` implementations.
        """
        for sut in await self.get_suts():
            entities = await self.get_entities()
            assert len(sut) == 0
            sut.add(*entities)
            assert len(sut) == len(entities)

    async def test___iter__(self) -> None:
        """
        Tests :py:meth:`betty.model.collections.EntityCollection.__iter__` implementations.
        """
        for sut in await self.get_suts():
            entities = await self.get_entities()
            assert list(iter(sut)) == []
            sut.add(*entities)
            assert list(iter(sut)) == list(entities)

    async def test_clear(self) -> None:
        """
        Tests :py:meth:`betty.model.collections.EntityCollection.__iter__` implementations.
        """
        for sut in await self.get_suts():
            entities = await self.get_entities()
            sut.add(*entities)
            sut.clear()
            assert list(sut) == []

    async def test_replace(self) -> None:
        """
        Tests :py:meth:`betty.model.collections.EntityCollection.replace` implementations.
        """
        for sut in await self.get_suts():
            entities = await self.get_entities()
            first = entities[0]
            others = entities[1:]
            sut.add(first)
            sut.replace(*others)
            assert list(sut) == list(others)

    async def test___getitem____by_index(self) -> None:
        """
        Tests :py:meth:`betty.model.collections.EntityCollection.__getitem__` implementations.
        """
        for sut in await self.get_suts():
            entities = await self.get_entities()
            sut.add(*entities)
            assert sut[0] == entities[0]

    async def test___getitem____by_indices(self) -> None:
        """
        Tests :py:meth:`betty.model.collections.EntityCollection.__getitem__` implementations.
        """
        for sut in await self.get_suts():
            entities = await self.get_entities()
            sut.add(*entities)
            assert list(sut[0:1:1]) == list(entities[0:1:1])
            assert list(sut[1::1]) == list(entities[1::1])

    async def test_get_mutable_instances(self) -> None:
        """
        Tests :py:meth:`betty.model.collections.EntityCollection.get_mutable_instances` implementations.
        """
        for sut in await self.get_suts():
            entities = await self.get_entities()
            sut.add(*entities)
            sut.immutable()
            for entity in entities:
                assert entity.is_immutable



"""
Provide localizable configuration.
"""

from contextlib import suppress
from typing import Self, final, overload, cast, TypeVar

from typing_extensions import override

from betty.assertion import assert_len
from betty.config import Configuration
from betty.locale import UNDETERMINED_LOCALE
from betty.locale.localizable import (
    ShorthandStaticTranslations,
    StaticTranslationsLocalizable,
)
from betty.locale.localizable.assertion import assert_static_translations
from betty.serde.dump import Dump

_T = TypeVar("_T")


@final
class StaticTranslationsLocalizableConfiguration(
    Configuration, StaticTranslationsLocalizable
):
    """
    Provide configuration for a :py:class:`betty.locale.localizable.Localizable`.

    Read more at :doc:`multiple translations </usage/configuration/static-translations-localizable>`.
    """

    @override
    def load(self, dump: Dump) -> None:
        self._translations.clear()

        translations = assert_static_translations()(dump)
        assert_len(minimum=1 if self._required else 0)(translations)
        for locale, translation in translations.items():
            self[locale] = translation

    @override
    def dump(self) -> Dump:
        translation_count = len(self._translations)
        if translation_count == 0:
            return {}
        elif translation_count == 1:
            with suppress(KeyError):
                return self._translations[UNDETERMINED_LOCALE]
        return dict(self._translations)


class _StaticTranslationsLocalizableConfigurationAttr:
    _required: bool

    def __init__(self, attr_name: str):
        self._attr_name = f"_{attr_name}"

    @overload
    def __get__(self, instance: None, owner: type[object]) -> Self:
        pass

    @overload
    def __get__(
        self, instance: _T, owner: type[_T]
    ) -> StaticTranslationsLocalizableConfiguration:
        pass

    def __get__(
        self, instance: object | None, owner: type[object]
    ) -> StaticTranslationsLocalizableConfiguration | Self:
        if instance is None:
            return self  # type: ignore[return-value]
        try:
            return cast(
                "StaticTranslationsLocalizableConfiguration",
                getattr(instance, self._attr_name),
            )
        except AttributeError:
            value = StaticTranslationsLocalizableConfiguration(required=self._required)
            setattr(instance, self._attr_name, value)
            return value

    def __set__(self, instance: object, value: ShorthandStaticTranslations) -> None:
        self.__get__(instance, type(instance)).replace(value)


@final
class RequiredStaticTranslationsLocalizableConfigurationAttr(
    _StaticTranslationsLocalizableConfigurationAttr
):
    """
    An instance attribute that contains :py:class:`betty.locale.localizable.config.StaticTranslationsLocalizableConfiguration`.
    """

    _required = True


@final
class OptionalStaticTranslationsLocalizableConfigurationAttr(
    _StaticTranslationsLocalizableConfigurationAttr
):
    """
    An instance attribute that contains :py:class:`betty.locale.localizable.config.StaticTranslationsLocalizableConfiguration`.
    """

    _required = False

    def __delete__(self, instance: object) -> None:
        self.__get__(instance, type(instance)).replace({})



"""
Provide copyright notices.
"""

from abc import abstractmethod

from betty.locale.localizable import Localizable
from betty.mutability import Mutable
from betty.plugin import Plugin, PluginRepository
from betty.plugin.entry_point import EntryPointPluginRepository


class CopyrightNotice(Mutable, Plugin):
    """
    A copyright notice.

    Read more about :doc:`/development/plugin/copyright-notice`.

    To test your own subclasses, use :py:class:`betty.test_utils.copyright_notice.CopyrightNoticeTestBase`.
    """

    @property
    @abstractmethod
    def summary(self) -> Localizable:
        """
        The copyright summary.
        """
        pass

    @property
    @abstractmethod
    def text(self) -> Localizable:
        """
        The full copyright text.
        """
        pass

    @property
    def url(self) -> Localizable | None:
        """
        The URL to an external human-readable resource with more information about this copyright.
        """
        return None


COPYRIGHT_NOTICE_REPOSITORY: PluginRepository[CopyrightNotice] = (
    EntryPointPluginRepository("betty.copyright_notice")
)
"""
The copyright notice plugin repository.

Read more about :doc:`/development/plugin/copyright-notice`.
"""



"""
Provide utilities for concurrent programming.
"""

import asyncio
import multiprocessing
import threading
import time
from abc import ABC, abstractmethod
from asyncio import sleep
from collections.abc import Hashable
from ctypes import c_longdouble
from math import floor
from multiprocessing.managers import SyncManager
from types import TracebackType
from typing import final, MutableMapping, TypeVar, Self, Union, TypeAlias

from typing_extensions import override

from betty.typing import processsafe
from betty.warnings import deprecate

_KeyT = TypeVar("_KeyT")
_ValueT = TypeVar("_ValueT")


class Lock(ABC):
    """
    Provide an asynchronous lock.
    """

    async def __aenter__(self):
        await self.acquire()

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        await self.release()

    @abstractmethod
    async def acquire(self, *, wait: bool = True) -> bool:
        """
        Acquire the lock.
        """
        pass

    @abstractmethod
    async def release(self) -> None:
        """
        Release the lock.
        """
        pass


Acquirable: TypeAlias = Union[threading.Lock, threading.Semaphore]


async def asynchronize_acquire(acquirable: Acquirable, *, wait: bool = True) -> bool:
    """
    Acquire a synchronous lock or semaphore asynchronously.
    """
    while not acquirable.acquire(blocking=False):
        if not wait:
            return False
        # Sleeping for zero seconds does not actually sleep, but gives the event
        # loop a chance to progress other tasks while we wait for another chance
        # to acquire the acquirable.
        await sleep(0)
    return True


@final
class AsynchronizedLock(Lock):
    """
    Make a synchronous (blocking) lock asynchronous (non-blocking).
    """

    __slots__ = "_lock"

    def __init__(self, lock: threading.Lock):
        self._lock = lock

    @property
    def lock(self) -> threading.Lock:
        """
        The underlying, synchronous lock.
        """
        return self._lock

    @override
    async def acquire(self, *, wait: bool = True) -> bool:
        return await asynchronize_acquire(self._lock, wait=wait)

    @override
    async def release(self) -> None:
        self._lock.release()

    @classmethod
    def threading(cls) -> Self:
        """
        Create a new thread-safe, asynchronous lock.
        """
        return cls(threading.Lock())


class Semaphore(ABC):
    """
    An asynchronous semaphore.
    """

    async def __aenter__(self):
        await self.acquire()

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        await self.release()

    @abstractmethod
    async def acquire(self, *, wait: bool = True) -> bool:
        """
        Acquire the semaphore.
        """
        pass

    @abstractmethod
    async def release(self, n: int = 1) -> None:
        """
        Release the semaphore.
        """
        pass


@final
class AsynchronizedSemaphore(Semaphore):
    """
    Make a synchronous (blocking) semaphore asynchronous (non-blocking).
    """

    __slots__ = "_semaphore"

    def __init__(self, semaphore: threading.Semaphore):
        self._semaphore = semaphore

    @property
    def semaphore(self) -> threading.Semaphore:
        """
        The underlying, synchronous semaphore.
        """
        return self._semaphore

    @override
    async def acquire(self, *, wait: bool = True) -> bool:
        return await asynchronize_acquire(self._semaphore, wait=wait)

    @override
    async def release(self, n: int = 1) -> None:
        self._semaphore.release(n)

    @classmethod
    def threading(cls, n: int = 1) -> Self:
        """
        Create a new thread-safe, asynchronous semaphore.
        """
        return cls(threading.Semaphore(n))


@final
@processsafe
class RateLimiter:
    """
    Rate-limit operations.

    This class implements the `Token Bucket algorithm <https://en.wikipedia.org/wiki/Token_bucket>`_.
    """

    def __init__(
        self, maximum: int, period: int = 1, *, manager: SyncManager | None = None
    ):
        manager = ensure_manager(manager)
        self._lock = AsynchronizedLock(manager.Lock())
        self._maximum = maximum
        self._period = period
        self._available = manager.Value(c_longdouble, maximum)
        # A Token Bucket fills as time passes. However, we want callers to be able to start
        # using the limiter immediately, so we 'preload' the first's period's tokens, and
        # set the last added time to the end of the first period. This ensures there is no
        # needless waiting if the number of tokens consumed in total is less than the limit
        # per period.
        self._last_add = manager.Value(c_longdouble, time.monotonic() + self._period)

    def _add_tokens(self):
        now = time.monotonic()
        elapsed = now - self._last_add.value
        added = elapsed * self._maximum
        possibly_available = floor(self._available.value + added)
        if possibly_available > 0:
            self._available.value = min(possibly_available, self._maximum)
            self._last_add.value = now

    async def __aenter__(self) -> None:
        await self.wait()

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        return

    async def is_available(self) -> bool:
        """
        Whether an operation may be performed (again).
        """
        async with self._lock:
            self._add_tokens()
            return self._available.value != 0

    async def wait(self) -> None:
        """
        Wait until an operation may be performed (again).
        """
        async with self._lock:
            while self._available.value < 1:
                self._add_tokens()
                if self._available.value < 1:
                    await asyncio.sleep(0)
            self._available.value -= 1


class _Transaction(Lock):
    def __init__(
        self,
        transaction_id: Hashable,
        ledger_lock: Lock,
        ledger: MutableMapping[Hashable, bool],
    ):
        self._transaction_id = transaction_id
        self._ledger_lock = ledger_lock
        self._ledger = ledger

    @override
    async def acquire(self, *, wait: bool = True) -> bool:
        if wait:
            while True:
                async with self._ledger_lock:
                    if self._can_acquire():
                        return self._acquire()
                await sleep(0)
        else:
            async with self._ledger_lock:
                if self._can_acquire():
                    return self._acquire()
                return False

    def _can_acquire(self) -> bool:
        try:
            return not self._ledger[self._transaction_id]
        except KeyError:
            self._ledger[self._transaction_id] = False
            return True

    def _acquire(self) -> bool:
        self._ledger[self._transaction_id] = True
        return True

    @override
    async def release(self) -> None:
        self._ledger[self._transaction_id] = False


@processsafe
class Ledger:
    """
    Lazily create locks by keeping a ledger.

    The ledger lock is released once a transaction lock is acquired.
    """

    def __init__(self, ledger_lock: Lock, *, manager: SyncManager | None = None):
        manager = ensure_manager(manager)
        self._ledger_lock = ledger_lock
        self._ledger: MutableMapping[Hashable, bool] = manager.dict()

    def ledger(self, transaction_id: Hashable) -> Lock:
        """
        Ledger a new lock for the given transaction ID.
        """
        return _Transaction(transaction_id, self._ledger_lock, self._ledger)


def ensure_manager(manager: SyncManager | None, *, stacklevel: int = 1) -> SyncManager:
    """
    Ensure that a value is a multiprocessing manager.
    """
    if manager:
        return manager
    deprecate(
        "Not providing a multiprocessing manager is deprecated as of Betty 0.4.10.",
        stacklevel=stacklevel,
    )
    return multiprocessing.Manager()



"""
Provide the Serve API to serve resources within the application.
"""

from __future__ import annotations

import contextlib
import logging
import threading
import webbrowser
from abc import ABC, abstractmethod
from asyncio import to_thread
from http.client import HTTPConnection
from http.server import SimpleHTTPRequestHandler, HTTPServer
from io import StringIO
from pathlib import Path
from typing import Any, TYPE_CHECKING, Self
from typing import final
from urllib.parse import urlparse

from aiofiles.os import makedirs, symlink
from aiofiles.tempfile import TemporaryDirectory, AiofilesContextManagerTempDir
from typing_extensions import override

from betty.error import UserFacingError
from betty.functools import Do
from betty.locale.localizable import _
from betty.project.factory import ProjectDependentFactory

if TYPE_CHECKING:
    from betty.locale.localizer import Localizer
    from betty.project import Project
    from types import TracebackType

DEFAULT_PORT = 8000


class ServerNotStartedError(RuntimeError):
    """
    Raised when a web server has not (fully) started yet.
    """

    pass  # pragma: no cover


class NoPublicUrlBecauseServerNotStartedError(ServerNotStartedError):
    """
    A public URL is not yet available because the server has not (fully) started yet.
    """

    def __init__(self):
        super().__init__(
            "Cannot get the public URL for a server that has not started yet."
        )


class OsError(UserFacingError, OSError):
    """
    Raised for I/O errors.
    """

    pass  # pragma: no cover


class Server(ABC):
    """
    Provide a (development) web server.
    """

    def __init__(self, localizer: Localizer):
        self._localizer = localizer

    @abstractmethod
    async def start(self) -> None:  # noqa B027
        """
        Start the server.
        """
        pass

    async def show(self) -> None:
        """
        Show the served site to the user.
        """
        logging.getLogger(__name__).info(
            self._localizer._("Serving your site at {url}...").format(
                url=self.public_url,
            )
        )
        webbrowser.open_new_tab(self.public_url)

    @abstractmethod
    async def stop(self) -> None:  # noqa B027
        """
        Stop the server.
        """
        pass

    @property
    @abstractmethod
    def public_url(self) -> str:
        """
        The server's public URL.
        """
        pass

    async def __aenter__(self) -> Server:
        await self.start()
        try:
            await self.assert_available()
        except BaseException:
            await self.stop()
            raise
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        await self.stop()

    async def assert_available(self) -> None:
        """
        Assert that this server is available.
        """
        try:
            await Do[Any, None](self._assert_available).until()
        except Exception as error:
            raise UserFacingError(
                _("The server at {url} was unreachable after starting.").format(
                    url=self.public_url
                )
            ) from error

    async def _assert_available(self) -> None:
        await to_thread(self.__assert_available)

    def __assert_available(self) -> None:
        url = urlparse(self.public_url)
        connection = HTTPConnection(url.netloc)
        connection.request("GET", url.path)
        response = connection.getresponse()
        assert 400 > response.status >= 200


class ProjectServer(ProjectDependentFactory, Server):
    """
    A web server for a Betty project.
    """

    def __init__(self, localizer: Localizer, project: Project) -> None:
        super().__init__(localizer)
        self._project = project

    @override
    @classmethod
    async def new_for_project(cls, project: Project) -> Self:
        return cls(await project.app.localizer, project)


@final
class _BuiltinServerRequestHandler(SimpleHTTPRequestHandler):
    @override
    def end_headers(self) -> None:
        self.send_header("Cache-Control", "no-cache")
        super().end_headers()


@final
class BuiltinServer(Server):
    """
    A built-in server for a WWW directory.
    """

    def __init__(
        self,
        www_directory_path: Path,
        *,
        root_path: str | None = None,
        localizer: Localizer,
    ) -> None:
        super().__init__(localizer)
        self._www_directory_path = www_directory_path
        self._root_path = root_path
        self._http_server: HTTPServer | None = None
        self._port: int | None = None
        self._thread: threading.Thread | None = None
        self._temporary_root_directory: AiofilesContextManagerTempDir | None = None

    @override
    async def start(self) -> None:
        if self._root_path:
            # To mimic the root path, symlink the project's WWW directory into a temporary
            # directory, so we do not have to make changes to any existing files.
            self._temporary_root_directory = TemporaryDirectory()
            temporary_root_directory_path = Path(
                await self._temporary_root_directory.__aenter__()
            )
            temporary_www_directory = temporary_root_directory_path
            for root_path_component in self._root_path.split("/"):
                temporary_www_directory /= root_path_component
            if temporary_www_directory != temporary_root_directory_path:
                temporary_www_directory.parent.mkdir(parents=True, exist_ok=True)
                await symlink(self._www_directory_path, temporary_www_directory)
            www_directory_path = temporary_root_directory_path
        else:
            www_directory_path = self._www_directory_path
        logging.getLogger(__name__).info(
            self._localizer._("Starting Python's built-in web server...")
        )
        for self._port in range(DEFAULT_PORT, 65535):
            with contextlib.suppress(OSError):
                self._http_server = HTTPServer(
                    ("", self._port),
                    lambda request,
                    client_address,
                    server: _BuiltinServerRequestHandler(
                        request,
                        client_address,
                        server,
                        directory=str(www_directory_path),
                    ),
                )
                break
        if self._http_server is None:
            raise OsError(_("Cannot find an available port to bind the web server to."))
        self._thread = threading.Thread(target=self._serve)
        self._thread.start()

    @override
    @property
    def public_url(self) -> str:
        if self._port is not None:
            url = f"http://localhost:{self._port}"
            if self._root_path:
                url = f"{url}/{self._root_path}"
            return url
        raise NoPublicUrlBecauseServerNotStartedError()

    def _serve(self) -> None:
        with contextlib.redirect_stderr(StringIO()):
            assert self._http_server
            self._http_server.serve_forever()

    @override
    async def stop(self) -> None:
        if self._http_server is not None:
            self._http_server.shutdown()
            self._http_server.server_close()
        if self._thread is not None:
            self._thread.join()
        if self._temporary_root_directory is not None:
            await self._temporary_root_directory.__aexit__(None, None, None)


@final
class BuiltinProjectServer(ProjectServer):
    """
    A built-in server for a Betty project.
    """

    def __init__(self, localizer: Localizer, project: Project) -> None:
        super().__init__(localizer, project)
        self._server = BuiltinServer(
            project.configuration.www_directory_path,
            root_path=project.configuration.root_path,
            localizer=localizer,
        )

    @override
    @property
    def public_url(self) -> str:
        return self._server.public_url

    @override
    async def start(self) -> None:
        await makedirs(self._project.configuration.www_directory_path, exist_ok=True)
        await self._server.start()

    @override
    async def stop(self) -> None:
        await self._server.stop()



"""
Provide the Dispatch API.
"""

from __future__ import annotations

from asyncio import gather
from collections import defaultdict
from typing import (
    Sequence,
    Callable,
    Awaitable,
    TypeVar,
    TypeAlias,
    TYPE_CHECKING,
    final,
)

if TYPE_CHECKING:
    from collections.abc import MutableMapping, MutableSequence, Mapping


class Event:
    """
    An event that can be dispatched and handled.

    For each unique 'conceptual' event (a thing that happens while using Betty),
    a subclass **MUST** be created for that specific event type. Instances of
    these subclasses are dispatched and handled.
    """

    pass


_EventT = TypeVar("_EventT", bound=Event)
EventHandler: TypeAlias = Callable[[_EventT], Awaitable[None]]


class _EventHandlerRegistry:
    """
    Manage event handlers.
    """

    def __init__(self):
        self._handlers: MutableMapping[
            type[Event], MutableSequence[Sequence[EventHandler[Event]]]
        ] = defaultdict(list)

    def add_handler(self, event_type: type[_EventT], *handlers: EventHandler[_EventT]):
        """
        Add a batch of one or more event handlers.

        All handlers of a batch are invoked concurrently.
        """
        self._handlers[event_type].append(
            handlers  # type: ignore[arg-type]
        )

    def add_registry(self, event_handler_registry: EventHandlerRegistry) -> None:
        """
        Add another registry to this one.
        """
        for (
            event_type,
            event_type_handler_batches,
        ) in event_handler_registry.handlers.items():
            for event_type_handler_batch in event_type_handler_batches:
                self.add_handler(event_type, *event_type_handler_batch)


@final
class EventHandlerRegistry(_EventHandlerRegistry):
    """
    Manage event handlers.
    """

    @property
    def handlers(
        self,
    ) -> Mapping[type[Event], Sequence[Sequence[EventHandler[Event]]]]:
        """
        The registered event handlers.
        """
        return self._handlers


@final
class EventDispatcher(_EventHandlerRegistry):
    """
    Dispatch events to event handlers.
    """

    async def dispatch(self, event: Event) -> None:
        """
        Dispatch an event.
        """
        for handler_batch in self._handlers[type(event)]:
            await gather(*(handler(event) for handler in handler_batch))



"""
Provide licenses.
"""

from abc import abstractmethod

from betty.locale.localizable import Localizable
from betty.mutability import Mutable
from betty.plugin import Plugin, PluginRepository
from betty.plugin.entry_point import EntryPointPluginRepository


class License(Mutable, Plugin):
    """
    A license.

    Read more about :doc:`/development/plugin/license`.

    To test your own subclasses, use :py:class:`betty.test_utils.license.LicenseTestBase`.
    """

    @property
    @abstractmethod
    def summary(self) -> Localizable:
        """
        The license summary.
        """
        pass

    @property
    @abstractmethod
    def text(self) -> Localizable:
        """
        The full license text.
        """
        pass

    @property
    def url(self) -> Localizable | None:
        """
        The URL to an external human-readable resource with more information about this license.
        """
        return None


LICENSE_REPOSITORY: PluginRepository[License] = EntryPointPluginRepository(
    "betty.license"
)
"""
The license plugin repository.

Read more about :doc:`/development/plugin/license`.
"""



"""
Integrates the plugin API with `distribution packages <https://packaging.python.org/en/latest/glossary/#term-Distribution-Package>`_.
"""

from collections.abc import Sequence
from importlib import metadata
from typing import Generic, TypeVar, final

from betty.json.schema import Schema
from betty.plugin import Plugin
from betty.plugin.lazy import LazyPluginRepositoryBase

_PluginT = TypeVar("_PluginT", bound=Plugin)


@final
class EntryPointPluginRepository(LazyPluginRepositoryBase[_PluginT], Generic[_PluginT]):
    """
    Discover plugins defined as distribution package `entry points <https://packaging.python.org/en/latest/specifications/entry-points/>`_.

    If you are developing a plugin for an existing plugin type that uses entry points, you'll have
    to add that plugin to your package metadata. For example, for a plugin type

    - whose entry point group is ``your-plugin-group``
    - with a plugin class ``MyPlugin`` in the module ``my_package.my_module``
    - and a plugin ID ``my-package-plugin``:

    .. code-block:: toml

        [project.entry-points.'your-plugin-group']
        'my-package-plugin' = 'my_package.my_module:MyPlugin'
    """

    def __init__(
        self, entry_point_group: str, *, schema_template: Schema | None = None
    ):
        super().__init__(schema_template=schema_template)
        self._entry_point_group = entry_point_group

    async def _load_plugins(self) -> Sequence[type[_PluginT]]:
        return [
            entry_point.load()
            for entry_point in metadata.entry_points(
                group=self._entry_point_group,
            )
        ]



"""
Providing typing utilities.
"""

from __future__ import annotations

from typing import TypeVar, TypeAlias, final, Any

from typing_extensions import TypeIs

from betty.docstring import append

_T = TypeVar("_T")


def _should_mark(target: Any, key: str) -> bool:
    attr_name = f"_betty_typing_{key}"
    if hasattr(target, attr_name):
        return False
    setattr(target, attr_name, True)
    return True


def _internal(target: _T) -> _T:
    if _should_mark(target, "internal"):
        target.__doc__ = append(
            target.__doc__ or "",
            "This is internal. It **MAY** be used anywhere in Betty's source code, but **MUST NOT** be used by third-party code.",
        )
    return target


@_internal
def internal(target: _T) -> _T:
    """
    Mark a target as internal to Betty.

    Anything decorated with ``@internal`` MAY be used anywhere in Betty's source code,
    but MUST be considered private by third-party code.
    """
    return _internal(target)


@internal
def public(target: _T) -> _T:
    """
    Mark a target as publicly usable.

    This is intended for items nested inside something marked with :py:func:`betty.typing.internal`,
    such as class attributes: third-party code **SHOULD NOT** use a class marked ``@internal``
    directly, but **MAY** use any of its attributes that are marked ``@public``.
    """
    return target


def private(target: _T) -> _T:
    """
    Mark a target as private to its containing scope.

    This is intended for items that cannot be marked private by prefixing their names with an underscore.
    """
    if _should_mark(target, "private"):
        target.__doc__ = append(
            target.__doc__ or "",
            "This is private. It **MUST NOT** be used anywhere outside its containing scope.",
        )
    return target


def pickleable(target: _T) -> _T:
    """
    Mark a target as pickleable.
    """
    if _should_mark(target, "pickleable"):
        target.__doc__ = append(
            target.__doc__ or "",
            "This can be pickled.",
        )
    return target


def unpickleable(target: _T) -> _T:
    """
    Mark a target as unpickleable.
    """
    if _should_mark(target, "unpickleable"):
        target.__doc__ = append(
            target.__doc__ or "",
            "This can NOT be pickled, and MUST NOT be used between different processes.",
        )

        def _reduce_ex(*_) -> None:
            raise RuntimeError(f"{target} is not pickleable.")

        target.__reduce_ex__ = _reduce_ex  # type: ignore[assignment, method-assign]
    return target


def threadsafe(target: _T) -> _T:
    """
    Mark a target as thread-safe.
    """
    if _should_mark(target, "threadsafe"):
        target.__doc__ = append(
            target.__doc__ or "",
            "This is thread-safe, which means you can safely use this between different threads.",
        )
    return target


def processsafe(target: _T) -> _T:
    """
    Mark a target as process-safe.
    """
    if _should_mark(target, "processsafe"):
        target = pickleable(target)
        target = threadsafe(target)
        target.__doc__ = append(
            target.__doc__ or "",
            "This is process-safe, which means you can safely use this between different processes.",
        )
    return target


@final
class Void:
    """
    A sentinel that describes the absence of a value.

    Using this sentinel allows for actual values to be ``None``. Like ``None``,
    ``Void`` is only ever used through its type, and never instantiated.
    """

    def __new__(cls):  # pragma: no cover  # noqa D102
        raise RuntimeError("The Void sentinel cannot be instantiated.")


Voidable: TypeAlias = _T | type[Void]


def not_void(value: _T | type[Void]) -> TypeIs[_T]:
    """
    Test that a value is not :py:class:`betty.typing.Void`.
    """
    return value is not Void



"""
Provide the Locale API.
"""

from __future__ import annotations

from typing import Sequence, TypeAlias

from babel import Locale, negotiate_locale as babel_negotiate_locale
from babel.core import UnknownLocaleError
from langcodes import Language

from betty import fs
from betty.json.schema import String

_LOCALE_DIRECTORY_PATH = fs.ASSETS_DIRECTORY_PATH / "locale"

NO_LINGUISTIC_CONTENT = "zxx"
UNDETERMINED_LOCALE = "und"
UNCODED_LOCALE = "mis"
MULTIPLE_LOCALES = "mul"
SPECIAL_LOCALES = (
    NO_LINGUISTIC_CONTENT,
    UNDETERMINED_LOCALE,
    UNCODED_LOCALE,
    MULTIPLE_LOCALES,
)

DEFAULT_LOCALE = "en-US"


def merge_locales(*locales: str) -> str:
    """
    Merge locales into a single locale.
    """
    unique_locales = set(locales)
    if len(unique_locales) == 0:
        return NO_LINGUISTIC_CONTENT
    elif len(unique_locales) == 1:
        return next(iter(unique_locales))
    # Strip locales without linguistic content.
    if NO_LINGUISTIC_CONTENT in unique_locales:
        return merge_locales(*(unique_locales - {NO_LINGUISTIC_CONTENT}))
    return MULTIPLE_LOCALES


def to_babel_identifier(locale: Localey) -> str:
    """
    Convert a locale or locale metadata to a Babel locale identifier.

    :raises ValueError:
    """
    if isinstance(locale, Locale):
        return str(locale)
    language_data = Language.get(locale)
    return "_".join(
        part
        for part in [
            language_data.language,
            language_data.script,
            language_data.territory,
        ]
        if part
    )


def to_locale(locale: Localey) -> str:
    """
    Ensure that a locale or locale metadata is a locale.
    """
    if isinstance(locale, str):
        return locale
    return "-".join(
        part
        for part in [
            locale.language,
            locale.script,
            locale.territory,
        ]
        if part
    )


Localey: TypeAlias = str | Locale


def get_data(locale: Localey) -> Locale:
    """
    Get locale metadata.

    :raises betty.locale.InvalidLocale: Raised if the given identifier is not a valid locale.
    :raises betty.locale.LocaleNotFoundError: Raised if the given locale cannot be found.
    """
    if isinstance(locale, Locale):
        return locale
    try:
        return Locale.parse(to_babel_identifier(locale))
    except ValueError:
        from betty.locale.error import InvalidLocale

        raise InvalidLocale(locale) from None
    except UnknownLocaleError:
        from betty.locale.error import LocaleNotFound

        raise LocaleNotFound(locale) from None


def get_display_name(
    locale: Localey, display_locale: Localey | None = None
) -> str | None:
    """
    Return a locale's human-readable display name.
    """
    locale_data = get_data(locale)
    return locale_data.get_display_name(
        get_data(display_locale) if display_locale else locale_data
    )


def negotiate_locale(
    preferred_locales: Localey | Sequence[Localey], available_locales: Sequence[Localey]
) -> Locale | None:
    """
    Negotiate the preferred locale from a sequence.
    """
    if isinstance(preferred_locales, (str, Locale)):
        preferred_locales = [preferred_locales]
    return _negotiate_locale(
        [to_babel_identifier(locale) for locale in preferred_locales],
        {to_babel_identifier(locale) for locale in available_locales},
        False,
    )


def _negotiate_locale(
    preferred_locale_babel_identifiers: Sequence[str],
    available_locale_babel_identifiers: set[str],
    root: bool,
) -> Locale | None:
    negotiated_locale = babel_negotiate_locale(
        preferred_locale_babel_identifiers, available_locale_babel_identifiers
    )
    if negotiated_locale is not None:
        return Locale.parse(negotiated_locale)
    if not root:
        return _negotiate_locale(
            [
                (
                    babel_identifier.split("_")[0]
                    if "_" in babel_identifier
                    else babel_identifier
                )
                for babel_identifier in preferred_locale_babel_identifiers
            ],
            available_locale_babel_identifiers,
            True,
        )
    return None


class LocaleSchema(String):
    """
    The JSON Schema for locales.
    """

    def __init__(self):
        super().__init__(
            def_name="locale",
            title="Locale",
            description="A BCP 47 locale identifier (https://www.ietf.org/rfc/bcp/bcp47.txt).",
        )



"""
Provide project configuration.
"""

from __future__ import annotations

from collections.abc import Mapping
from reprlib import recursive_repr
from typing import final, Self, Iterable, Any, TYPE_CHECKING, cast
from urllib.parse import urlparse

from typing_extensions import override

from betty.ancestry.event import Event
from betty.ancestry.event_type import EventType
from betty.ancestry.gender import Gender
from betty.ancestry.person import Person
from betty.ancestry.place import Place
from betty.ancestry.place_type import PlaceType
from betty.ancestry.presence_role import PresenceRole
from betty.ancestry.source import Source
from betty.assertion import (
    assert_record,
    RequiredField,
    assert_setattr,
    OptionalField,
    assert_str,
    assert_bool,
    assert_fields,
    assert_locale,
    assert_positive_number,
    assert_int,
    assert_path,
    assert_mapping,
    assert_none,
    assert_or,
)
from betty.assertion.error import AssertionFailed, AssertionFailedGroup, Key
from betty.config import Configuration
from betty.config.collections.mapping import (
    OrderedConfigurationMapping,
)
from betty.copyright_notice import CopyrightNotice
from betty.license import License
from betty.license.licenses import AllRightsReserved
from betty.locale import DEFAULT_LOCALE, UNDETERMINED_LOCALE
from betty.locale.localizable import _, ShorthandStaticTranslations, Localizable
from betty.locale.localizable.assertion import assert_static_translations
from betty.locale.localizable.config import (
    OptionalStaticTranslationsLocalizableConfigurationAttr,
    RequiredStaticTranslationsLocalizableConfigurationAttr,
)
from betty.machine_name import MachineName
from betty.machine_name import assert_machine_name
from betty.model import Entity
from betty.plugin import (
    ShorthandPluginBase,
    PluginRepository,
)
from betty.plugin.config import (
    PluginConfigurationPluginConfigurationMapping,
    PluginConfiguration,
    PluginConfigurationMapping,
    PluginInstanceConfiguration,
    PluginInstanceConfigurationMapping,
    PluginIdentifierKeyConfigurationMapping,
)
from betty.project.extension import Extension
from betty.repr import repr_instance
from betty.serde.format import Format, format_for, FORMAT_REPOSITORY
from betty.user import UserFacing

if TYPE_CHECKING:
    from betty.serde.dump import Dump, DumpMapping
    from collections.abc import Sequence
    from pathlib import Path


#: The default age by which people are presumed dead.
#: This is based on `Jeanne Louise Calment <https://www.guinnessworldrecords.com/world-records/oldest-person/>`_ who is
#: the oldest verified person to ever have lived.
DEFAULT_LIFETIME_THRESHOLD = 123


@final
class ExtensionInstanceConfigurationMapping(
    PluginInstanceConfigurationMapping[Extension]
):
    """
    Configure a project's enabled extensions.
    """

    def enable(self, *extension_types: type[Extension] | MachineName) -> None:
        """
        Enable the given extensions.
        """
        for extension_type in extension_types:
            if extension_type not in self._configurations:
                self.append(PluginInstanceConfiguration(extension_type))


@final
class EntityTypeConfiguration(Configuration):
    """
    Configure a single entity type for a project.
    """

    def __init__(
        self,
        entity_type_id: MachineName | type[Entity],
        *,
        generate_html_list: bool = False,
    ):
        super().__init__()
        self._id = (
            entity_type_id
            if isinstance(entity_type_id, str)
            else entity_type_id.plugin_id()
        )
        self.generate_html_list = generate_html_list

    @property
    def id(self) -> MachineName:
        """
        The ID of the configured entity type.
        """
        return self._id

    @property
    def generate_html_list(self) -> bool:
        """
        Whether to generate listing web pages for entities of this type.
        """
        return self._generate_html_list

    @generate_html_list.setter
    def generate_html_list(self, generate_html_list: bool) -> None:
        self._generate_html_list = generate_html_list

    @override
    def load(self, dump: Dump) -> None:
        assert_record(
            RequiredField[Any, type[Entity]](
                "id", assert_machine_name() | assert_setattr(self, "_id")
            ),
            OptionalField(
                "generate_html_list",
                assert_bool() | assert_setattr(self, "generate_html_list"),
            ),
        )(dump)

    @override
    def dump(self) -> DumpMapping[Dump]:
        return {
            "id": self.id,
            "generate_html_list": self.generate_html_list,
        }

    async def validate(self, entity_type_repository: PluginRepository[Entity]) -> None:
        """
        Validate the configuration.
        """
        entity_type = await entity_type_repository.get(self.id)
        if self.generate_html_list and not issubclass(entity_type, UserFacing):
            raise AssertionFailed(
                _(
                    "Cannot generate pages for {entity_type}, because it is not a user-facing entity type."
                ).format(entity_type=entity_type.plugin_label())
            )


@final
class EntityTypeConfigurationMapping(
    PluginIdentifierKeyConfigurationMapping[Entity, EntityTypeConfiguration]
):
    """
    Configure the entity types for a project.
    """

    @override
    def _get_key(self, configuration: EntityTypeConfiguration) -> MachineName:
        return configuration.id

    @override
    def _load_key(self, item_dump: Dump, key_dump: str) -> Dump:
        assert isinstance(item_dump, Mapping)
        item_dump["id"] = key_dump
        return item_dump

    @override
    def _dump_key(self, item_dump: Dump) -> tuple[Dump, str]:
        assert isinstance(item_dump, Mapping)
        return item_dump, cast("str", item_dump.pop("id"))

    @override
    def _load_item(self, dump: Dump) -> EntityTypeConfiguration:
        # Use a dummy entity type for now to satisfy the initializer.
        # It will be overridden when loading the dump.
        configuration = EntityTypeConfiguration(Entity)
        configuration.load(dump)
        return configuration

    async def validate(self, entity_type_repository: PluginRepository[Entity]) -> None:
        """
        Validate the configuration.
        """
        with AssertionFailedGroup().assert_valid() as errors:
            for configuration in self.values():
                with errors.catch(Key(configuration.id)):
                    await configuration.validate(entity_type_repository)


@final
class LocaleConfiguration(Configuration):
    """
    Configure a single project locale.
    """

    def __init__(
        self,
        locale: str,
        *,
        alias: str | None = None,
    ):
        super().__init__()
        self._locale = locale
        if alias is not None and "/" in alias:
            raise AssertionFailed(_("Locale aliases must not contain slashes."))
        self._alias = alias

    @override  # type: ignore[callable-functiontype]
    @recursive_repr()
    def __repr__(self) -> str:
        return repr_instance(self, locale=self.locale, alias=self.alias)

    @property
    def locale(self) -> str:
        """
        An `IETF BCP 47 <https://tools.ietf.org/html/bcp47>`_ language tag.
        """
        return self._locale

    @property
    def alias(self) -> str:
        """
        A shorthand alias to use instead of the full language tag, such as when rendering URLs.
        """
        if self._alias is None:
            return self.locale
        return self._alias

    @alias.setter
    def alias(self, alias: str | None) -> None:
        self._alias = alias

    @override
    def load(self, dump: Dump) -> None:
        assert_record(
            RequiredField("locale", assert_locale() | assert_setattr(self, "_locale")),
            OptionalField(
                "alias",
                assert_or(assert_str() | assert_setattr(self, "alias"), assert_none()),
            ),
        )(dump)

    @override
    def dump(self) -> Dump:
        return {"locale": self.locale, "alias": self._alias}


@final
class LocaleConfigurationMapping(OrderedConfigurationMapping[str, LocaleConfiguration]):
    """
    Configure a project's locales.
    """

    def __init__(
        self,
        configurations: Iterable[LocaleConfiguration] | None = None,
    ):
        super().__init__(configurations)
        self._ensure_locale()

    @override
    def _post_remove(self, configuration: LocaleConfiguration) -> None:
        super()._post_remove(configuration)
        self._ensure_locale()

    def _ensure_locale(self) -> None:
        if len(self) == 0:
            self.append(LocaleConfiguration(DEFAULT_LOCALE))

    @override
    def replace(self, *configurations: LocaleConfiguration) -> None:
        # Prevent the events from being dispatched.
        self._configurations.clear()
        self.append(*configurations)
        self._ensure_locale()

    @override
    def _load_item(self, dump: Dump) -> LocaleConfiguration:
        item = LocaleConfiguration(UNDETERMINED_LOCALE)
        item.load(dump)
        return item

    @override
    def _get_key(self, configuration: LocaleConfiguration) -> str:
        return configuration.locale

    @property
    def default(self) -> LocaleConfiguration:
        """
        The default language.
        """
        return next(self.values())

    @property
    def multilingual(self) -> bool:
        """
        Whether the configuration is multilingual.
        """
        return len(self) > 1


class CopyrightNoticeConfiguration(PluginConfiguration):
    """
    Configuration to define :py:class:`betty.copyright_notice.CopyrightNotice` plugins.
    """

    summary = RequiredStaticTranslationsLocalizableConfigurationAttr("summary")
    text = RequiredStaticTranslationsLocalizableConfigurationAttr("text")

    def __init__(
        self,
        plugin_id: MachineName,
        label: ShorthandStaticTranslations,
        *,
        summary: ShorthandStaticTranslations,
        text: ShorthandStaticTranslations,
        description: ShorthandStaticTranslations | None = None,
    ):
        super().__init__(plugin_id, label, description=description)
        self.summary = summary
        self.text = text

    @override
    def load(self, dump: Dump) -> None:
        mapping = assert_mapping()(dump)
        assert_fields(
            RequiredField(
                "summary",
                assert_static_translations() | assert_setattr(self, "summary"),
            ),
            RequiredField(
                "text",
                assert_static_translations() | assert_setattr(self, "text"),
            ),
        )(mapping)
        del mapping["summary"]
        del mapping["text"]
        super().load(mapping)

    @override
    def dump(self) -> DumpMapping[Dump]:
        return {
            **super().dump(),
            "summary": self.summary.dump(),
            "text": self.text.dump(),
        }


class CopyrightNoticeConfigurationMapping(
    PluginConfigurationMapping[CopyrightNotice, CopyrightNoticeConfiguration]
):
    """
    A configuration mapping for copyright notices.
    """

    @override
    def _new_plugin(
        self, configuration: CopyrightNoticeConfiguration
    ) -> type[CopyrightNotice]:
        class _ProjectConfigurationCopyrightNotice(
            ShorthandPluginBase, CopyrightNotice
        ):
            _plugin_id = configuration.id
            _plugin_label = configuration.label
            _plugin_description = configuration.description

            @override
            @property
            def summary(self) -> Localizable:
                return configuration.summary

            @override
            @property
            def text(self) -> Localizable:
                return configuration.text

        return _ProjectConfigurationCopyrightNotice

    @override
    def _load_item(self, dump: Dump) -> CopyrightNoticeConfiguration:
        item = CopyrightNoticeConfiguration("-", "", summary="", text="")
        item.load(dump)
        return item

    @classmethod
    def _create_default_item(
        cls, configuration_key: str
    ) -> CopyrightNoticeConfiguration:
        return CopyrightNoticeConfiguration(configuration_key, {}, summary="", text="")


class LicenseConfiguration(PluginConfiguration):
    """
    Configuration to define :py:class:`betty.license.License` plugins.
    """

    summary = RequiredStaticTranslationsLocalizableConfigurationAttr("summary")
    text = RequiredStaticTranslationsLocalizableConfigurationAttr("text")

    def __init__(
        self,
        plugin_id: MachineName,
        label: ShorthandStaticTranslations,
        *,
        summary: ShorthandStaticTranslations,
        text: ShorthandStaticTranslations,
        description: ShorthandStaticTranslations | None = None,
    ):
        super().__init__(plugin_id, label, description=description)
        self.summary = summary
        self.text = text

    @override
    def load(self, dump: Dump) -> None:
        mapping = assert_mapping()(dump)
        assert_fields(
            RequiredField(
                "summary",
                assert_static_translations() | assert_setattr(self, "summary"),
            ),
            RequiredField(
                "text",
                assert_static_translations() | assert_setattr(self, "text"),
            ),
        )(mapping)
        del mapping["summary"]
        del mapping["text"]
        super().load(mapping)

    @override
    def dump(self) -> DumpMapping[Dump]:
        return {
            **super().dump(),
            "summary": self.summary.dump(),
            "text": self.text.dump(),
        }


class LicenseConfigurationMapping(
    PluginConfigurationMapping[License, LicenseConfiguration]
):
    """
    A configuration mapping for licenses.
    """

    @override
    def _new_plugin(self, configuration: LicenseConfiguration) -> type[License]:
        class _ProjectConfigurationLicense(ShorthandPluginBase, License):
            _plugin_id = configuration.id
            _plugin_label = configuration.label
            _plugin_description = configuration.description

            @override
            @property
            def summary(self) -> Localizable:
                return configuration.summary

            @override
            @property
            def text(self) -> Localizable:
                return configuration.text

        return _ProjectConfigurationLicense

    @override
    def _load_item(self, dump: Dump) -> LicenseConfiguration:
        item = LicenseConfiguration("-", "", summary="", text="")
        item.load(dump)
        return item

    @classmethod
    def _create_default_item(cls, configuration_key: str) -> LicenseConfiguration:
        return LicenseConfiguration(configuration_key, {}, summary="", text="")


class EventTypeConfigurationMapping(
    PluginConfigurationPluginConfigurationMapping[EventType]
):
    """
    A configuration mapping for event types.
    """

    @override
    def _new_plugin(self, configuration: PluginConfiguration) -> type[EventType]:
        class _ProjectConfigurationEventType(ShorthandPluginBase, EventType):
            _plugin_id = configuration.id
            _plugin_label = configuration.label
            _plugin_description = configuration.description

        return _ProjectConfigurationEventType


class PlaceTypeConfigurationMapping(
    PluginConfigurationPluginConfigurationMapping[PlaceType]
):
    """
    A configuration mapping for place types.
    """

    @override
    def _new_plugin(self, configuration: PluginConfiguration) -> type[PlaceType]:
        class _ProjectConfigurationPlaceType(ShorthandPluginBase, PlaceType):
            _plugin_id = configuration.id
            _plugin_label = configuration.label
            _plugin_description = configuration.description

        return _ProjectConfigurationPlaceType


class PresenceRoleConfigurationMapping(
    PluginConfigurationPluginConfigurationMapping[PresenceRole]
):
    """
    A configuration mapping for presence roles.
    """

    @override
    def _new_plugin(self, configuration: PluginConfiguration) -> type[PresenceRole]:
        class _ProjectConfigurationPresenceRole(ShorthandPluginBase, PresenceRole):
            _plugin_id = configuration.id
            _plugin_label = configuration.label
            _plugin_description = configuration.description

        return _ProjectConfigurationPresenceRole


class GenderConfigurationMapping(PluginConfigurationPluginConfigurationMapping[Gender]):
    """
    A configuration mapping for genders.
    """

    @override
    def _new_plugin(self, configuration: PluginConfiguration) -> type[Gender]:
        class _ProjectConfigurationGender(ShorthandPluginBase, Gender):
            _plugin_id = configuration.id
            _plugin_label = configuration.label
            _plugin_description = configuration.description

        return _ProjectConfigurationGender


@final
class ProjectConfiguration(Configuration):
    """
    Provide the configuration for a :py:class:`betty.project.Project`.
    """

    title = OptionalStaticTranslationsLocalizableConfigurationAttr("title")
    author = OptionalStaticTranslationsLocalizableConfigurationAttr("author")
    #: The project-wide :py:class:`betty.copyright_notice.CopyrightNotice` plugin to use.
    copyright_notice: PluginInstanceConfiguration
    #: The project-wide :py:class:`betty.license.License` plugin to use.
    license: PluginInstanceConfiguration

    def __init__(
        self,
        configuration_file_path: Path,
        *,
        available_formats: Sequence[type[Format]],
        url: str = "https://example.com",
        clean_urls: bool = False,
        title: ShorthandStaticTranslations = "Betty",
        author: ShorthandStaticTranslations | None = None,
        entity_types: Iterable[EntityTypeConfiguration] | None = None,
        event_types: Iterable[PluginConfiguration] | None = None,
        place_types: Iterable[PluginConfiguration] | None = None,
        presence_roles: Iterable[PluginConfiguration] | None = None,
        copyright_notice: PluginInstanceConfiguration | None = None,  # noqa A002
        copyright_notices: Iterable[CopyrightNoticeConfiguration] | None = None,
        license: PluginInstanceConfiguration | None = None,  # noqa A002
        licenses: Iterable[LicenseConfiguration] | None = None,
        genders: Iterable[PluginConfiguration] | None = None,
        extensions: Iterable[PluginInstanceConfiguration] | None = None,
        debug: bool = False,
        locales: Iterable[LocaleConfiguration] | None = None,
        lifetime_threshold: int = DEFAULT_LIFETIME_THRESHOLD,
        name: MachineName | None = None,
        logo: Path | None = None,
    ):
        from betty.copyright_notice.copyright_notices import ProjectAuthor

        super().__init__()
        self._available_formats = available_formats
        self._configuration_file_path = configuration_file_path
        self._name = name
        self._computed_name: str | None = None
        self._url = url
        self._clean_urls = clean_urls
        self.title = title
        if author:
            self.author = author
        self._entity_types = EntityTypeConfigurationMapping(
            entity_types
            or [
                EntityTypeConfiguration(Person, generate_html_list=True),
                EntityTypeConfiguration(Event, generate_html_list=True),
                EntityTypeConfiguration(Place, generate_html_list=True),
                EntityTypeConfiguration(Source, generate_html_list=True),
            ]
        )
        self.copyright_notice = copyright_notice or PluginInstanceConfiguration(
            ProjectAuthor
        )
        self._copyright_notices = CopyrightNoticeConfigurationMapping()
        if copyright_notices is not None:
            self._copyright_notices.append(*copyright_notices)
        self.license = license or PluginInstanceConfiguration(AllRightsReserved)
        self._licenses = LicenseConfigurationMapping()
        if licenses is not None:
            self._licenses.append(*licenses)
        self._event_types = EventTypeConfigurationMapping()
        if event_types is not None:
            self._event_types.append(*event_types)
        self._place_types = PlaceTypeConfigurationMapping()
        if place_types is not None:
            self._place_types.append(*place_types)
        self._presence_roles = PresenceRoleConfigurationMapping()
        if presence_roles is not None:
            self._presence_roles.append(*presence_roles)
        self._genders = GenderConfigurationMapping()
        if genders is not None:
            self._genders.append(*genders)
        self._extensions = ExtensionInstanceConfigurationMapping(extensions or ())
        self._debug = debug
        self._locales = LocaleConfigurationMapping(locales or ())
        self._lifetime_threshold = lifetime_threshold
        self._logo = logo

    @classmethod
    async def new(
        cls,
        configuration_file_path: Path,
        *,
        url: str = "https://example.com",
        clean_urls: bool = False,
        title: ShorthandStaticTranslations = "Betty",
        author: ShorthandStaticTranslations | None = None,
        entity_types: Iterable[EntityTypeConfiguration] | None = None,
        event_types: Iterable[PluginConfiguration] | None = None,
        place_types: Iterable[PluginConfiguration] | None = None,
        presence_roles: Iterable[PluginConfiguration] | None = None,
        copyright_notice: PluginInstanceConfiguration | None = None,  # noqa A002
        copyright_notices: Iterable[CopyrightNoticeConfiguration] | None = None,
        license: PluginInstanceConfiguration | None = None,  # noqa A002
        licenses: Iterable[LicenseConfiguration] | None = None,
        genders: Iterable[PluginConfiguration] | None = None,
        extensions: Iterable[PluginInstanceConfiguration] | None = None,
        debug: bool = False,
        locales: Iterable[LocaleConfiguration] | None = None,
        lifetime_threshold: int = DEFAULT_LIFETIME_THRESHOLD,
        name: MachineName | None = None,
        logo: Path | None = None,
    ) -> Self:
        """
        Create a new instance.
        """
        return cls(
            configuration_file_path,
            available_formats=await FORMAT_REPOSITORY.select(),
            url=url,
            clean_urls=clean_urls,
            title=title,
            author=author,
            entity_types=entity_types,
            event_types=event_types,
            place_types=place_types,
            presence_roles=presence_roles,
            copyright_notice=copyright_notice,
            copyright_notices=copyright_notices,
            license=license,
            licenses=licenses,
            genders=genders,
            extensions=extensions,
            debug=debug,
            locales=locales,
            lifetime_threshold=lifetime_threshold,
            name=name,
            logo=logo,
        )

    @property
    def configuration_file_path(self) -> Path:
        """
        The path to the configuration's file.
        """
        return self._configuration_file_path

    @configuration_file_path.setter
    def configuration_file_path(self, configuration_file_path: Path) -> None:
        self.assert_mutable()
        if configuration_file_path == self._configuration_file_path:
            return
        format_for(self._available_formats, configuration_file_path.suffix)
        self._configuration_file_path = configuration_file_path

    @property
    def name(self) -> MachineName | None:
        """
        The project's machine name.
        """
        return self._name

    @name.setter
    def name(self, name: MachineName) -> None:
        self.assert_mutable()
        self._name = assert_machine_name()(name)

    @property
    def project_directory_path(self) -> Path:
        """
        The project directory path.

        Betty will look for resources in this directory, and place generated artifacts there. It is expected
        that no other applications or projects share this same directory.
        """
        return self.configuration_file_path.parent

    @property
    def output_directory_path(self) -> Path:
        """
        The output directory path.
        """
        return self.project_directory_path / "output"

    @property
    def assets_directory_path(self) -> Path:
        """
        The :doc:`assets directory path </usage/assets>`.
        """
        return self.project_directory_path / "assets"

    @property
    def www_directory_path(self) -> Path:
        """
        The WWW directory path.
        """
        return self.output_directory_path / "www"

    def localize_www_directory_path(self, locale: str) -> Path:
        """
        Get the WWW directory path for a locale.
        """
        if self.locales.multilingual:
            return self.www_directory_path / self.locales[locale].alias
        return self.www_directory_path

    @property
    def url(self) -> str:
        """
        The project's public URL.
        """
        return self._url

    @url.setter
    def url(self, url: str) -> None:
        self.assert_mutable()
        url_parts = urlparse(url)
        if not url_parts.scheme:
            raise AssertionFailed(
                _("The URL must start with a scheme such as https:// or http://.")
            )
        if not url_parts.netloc:
            raise AssertionFailed(_("The URL must include a host."))
        self._url = f"{url_parts.scheme}://{url_parts.netloc}{url_parts.path}"

    @property
    def base_url(self) -> str:
        """
        The project's public URL's base URL.

        If the public URL is ``https://example.com``, the base URL is ``https://example.com``.
        If the public URL is ``https://example.com/my-ancestry-site``, the base URL is ``https://example.com``.
        If the public URL is ``https://my-ancestry-site.example.com``, the base URL is ``https://my-ancestry-site.example.com``.
        """
        url_parts = urlparse(self.url)
        return f"{url_parts.scheme}://{url_parts.netloc}"

    @property
    def root_path(self) -> str:
        """
        The project's public URL's root path.

        If the public URL is ``https://example.com``, the root path is an empty string.
        If the public URL is ``https://example.com/my-ancestry-site``, the root path is ``/my-ancestry-site``.
        """
        return urlparse(self.url).path.rstrip("/")

    @property
    def clean_urls(self) -> bool:
        """
        Whether to generate clean URLs such as ``/person/first-person`` instead of ``/person/first-person/index.html``.

        Generated artifacts will require web server that supports this.
        """
        return self._clean_urls

    @clean_urls.setter
    def clean_urls(self, clean_urls: bool) -> None:
        self.assert_mutable()
        self._clean_urls = clean_urls

    @property
    def locales(self) -> LocaleConfigurationMapping:
        """
        The available locales.
        """
        return self._locales

    @property
    def entity_types(self) -> EntityTypeConfigurationMapping:
        """
        The available entity types.
        """
        return self._entity_types

    @property
    def extensions(self) -> ExtensionInstanceConfigurationMapping:
        """
        Then extensions running within this application.
        """
        return self._extensions

    @property
    def debug(self) -> bool:
        """
        Whether to enable debugging for project jobs.

        This setting is disabled by default.

        Enabling this generally results in:

        - More verbose logging output
        - job artifacts (e.g. generated sites)
        """
        return self._debug

    @debug.setter
    def debug(self, debug: bool) -> None:
        self.assert_mutable()
        self._debug = debug

    @property
    def lifetime_threshold(self) -> int:
        """
        The lifetime threshold indicates when people are considered dead.

        This setting defaults to :py:const:`betty.project.config.DEFAULT_LIFETIME_THRESHOLD`.

        The value is an integer expressing the age in years over which people are
        presumed to have died.
        """
        return self._lifetime_threshold

    @lifetime_threshold.setter
    def lifetime_threshold(self, lifetime_threshold: int) -> None:
        self.assert_mutable()
        assert_positive_number()(lifetime_threshold)
        self._lifetime_threshold = lifetime_threshold

    @property
    def logo(self) -> Path | None:
        """
        The path to the logo.
        """
        return self._logo

    @logo.setter
    def logo(self, logo: Path | None) -> None:
        self.assert_mutable()
        self._logo = logo

    @property
    def copyright_notices(
        self,
    ) -> PluginConfigurationMapping[CopyrightNotice, CopyrightNoticeConfiguration]:
        """
        The :py:class:`betty.copyright_notice.CopyrightNotice` plugins created by this project.
        """
        return self._copyright_notices

    @property
    def licenses(
        self,
    ) -> PluginConfigurationMapping[License, LicenseConfiguration]:
        """
        The :py:class:`betty.license.License` plugins created by this project.
        """
        return self._licenses

    @property
    def event_types(self) -> PluginConfigurationMapping[EventType, PluginConfiguration]:
        """
        The event type plugins created by this project.
        """
        return self._event_types

    @property
    def place_types(self) -> PluginConfigurationMapping[PlaceType, PluginConfiguration]:
        """
        The place type plugins created by this project.
        """
        return self._place_types

    @property
    def presence_roles(
        self,
    ) -> PluginConfigurationMapping[PresenceRole, PluginConfiguration]:
        """
        The presence role plugins created by this project.
        """
        return self._presence_roles

    @property
    def genders(
        self,
    ) -> PluginConfigurationMapping[Gender, PluginConfiguration]:
        """
        The gender plugins created by this project.
        """
        return self._genders

    @override
    def load(self, dump: Dump) -> None:
        self.assert_mutable()
        assert_record(
            OptionalField(
                "name",
                assert_or(assert_str() | assert_setattr(self, "name"), assert_none()),
            ),
            RequiredField("url", assert_str() | assert_setattr(self, "url")),
            OptionalField("title", self.title.load),
            OptionalField("author", self.author.load),
            OptionalField(
                "logo",
                assert_or(assert_path() | assert_setattr(self, "logo"), assert_none()),
            ),
            OptionalField(
                "clean_urls",
                assert_bool() | assert_setattr(self, "clean_urls"),
            ),
            OptionalField("debug", assert_bool() | assert_setattr(self, "debug")),
            OptionalField(
                "lifetime_threshold",
                assert_int() | assert_setattr(self, "lifetime_threshold"),
            ),
            OptionalField("locales", self.locales.load),
            OptionalField("extensions", self.extensions.load),
            OptionalField("entity_types", self.entity_types.load),
            OptionalField("copyright_notice", self.copyright_notice.load),
            OptionalField("copyright_notices", self.copyright_notices.load),
            OptionalField("license", self.license.load),
            OptionalField("licenses", self.licenses.load),
            OptionalField("event_types", self.event_types.load),
            OptionalField("genders", self.genders.load),
            OptionalField("place_types", self.place_types.load),
            OptionalField("presence_roles", self.presence_roles.load),
        )(dump)

    @override
    def dump(self) -> DumpMapping[Dump]:
        dump: DumpMapping[Dump] = {
            "name": self.name,
            "url": self.url,
            "title": self.title.dump(),
            "clean_urls": self.clean_urls,
            "author": self.author.dump(),
            "logo": str(self._logo) if self._logo else None,
            "debug": self.debug,
            "lifetime_threshold": self.lifetime_threshold,
            "locales": self.locales.dump(),
            "extensions": self.extensions.dump(),
            "entity_types": self.entity_types.dump(),
            "copyright_notice": self.copyright_notice.dump(),
            "copyright_notices": self.copyright_notices.dump(),
            "license": self.license.dump(),
            "licenses": self.licenses.dump(),
            "event_types": self.event_types.dump(),
            "genders": self.genders.dump(),
            "place_types": self.place_types.dump(),
            "presence_roles": self.presence_roles.dump(),
        }
        return dump



"""
Fetch content from the internet.
"""

from __future__ import annotations

from typing import Mapping, TYPE_CHECKING

from typing_extensions import override

from betty.fetch import Fetcher, FetchResponse, FetchError
from betty.locale.localizable import static

if TYPE_CHECKING:
    from pathlib import Path


class StaticFetcher(Fetcher):
    """
    Return predefined (static) fetch responses.
    """

    def __init__(
        self,
        *,
        fetch_map: Mapping[str, FetchResponse] | None = None,
        fetch_file_map: Mapping[str, Path] | None = None,
    ):
        self._fetch_map = fetch_map or {}
        self._fetch_file_map = fetch_file_map or {}

    @override
    async def fetch(self, url: str) -> FetchResponse:
        try:
            return self._fetch_map[url]
        except KeyError:
            raise FetchError(static("")) from None

    @override
    async def fetch_file(self, url: str) -> Path:
        try:
            return self._fetch_file_map[url]
        except KeyError:
            raise FetchError(static("")) from None



from asyncio import BaseEventLoop
from multiprocessing import set_start_method
from typing import Any


_bootstrapped = False


if not _bootstrapped:
    _bootstrapped = True

    _original_shutdown_default_executor = BaseEventLoop.shutdown_default_executor

    async def _shutdown_default_executor(
        loop: BaseEventLoop, *args: Any, **kwargs: Any
    ) -> None:
        try:
            await _original_shutdown_default_executor(loop, *args, **kwargs)
        except RuntimeError as error:
            # Work around a bug in Python 3.12 that will randomly cause a RuntimeError with the
            # following message to be raised.
            if "can't create new thread at interpreter shutdown" not in str(error):
                raise

    BaseEventLoop.shutdown_default_executor = _shutdown_default_executor  # type: ignore[assignment, callable-functiontype, method-assign, misc]

    # Avoid `fork` so as not to start worker processes with unneeded resources.
    # Use `spawn`, which is the Python 3.14 default.
    set_start_method("spawn", force=True)



"""
Expand an ancestry by deriving additional data from existing data.
"""

from __future__ import annotations

from logging import getLogger
from typing import TYPE_CHECKING
from typing import final

from typing_extensions import override

from betty.ancestry import event_type
from betty.ancestry.event_type.event_types import DerivableEventType
from betty.deriver import Deriver as DeriverApi
from betty.locale.localizable import _
from betty.plugin import ShorthandPluginBase
from betty.project.extension import Extension
from betty.project.extension.privatizer import Privatizer
from betty.project.load import PostLoadAncestryEvent

if TYPE_CHECKING:
    from betty.plugin import PluginIdentifier
    from betty.event_dispatcher import EventHandlerRegistry


async def _derive_ancestry(event: PostLoadAncestryEvent) -> None:
    project = event.project
    localizer = await project.app.localizer
    logger = getLogger(__name__)
    logger.info(localizer._("Deriving..."))

    deriver = DeriverApi(
        project.ancestry,
        project.configuration.lifetime_threshold,
        project.event_type_repository,
        set(
            await event_type.EVENT_TYPE_REPOSITORY.select(
                DerivableEventType  # type: ignore[type-abstract]
            )
        ),
        localizer=localizer,
    )
    await deriver.derive()


@final
class Deriver(ShorthandPluginBase, Extension):
    """
    Expand an ancestry by deriving additional data from existing data.
    """

    _plugin_id = "deriver"
    _plugin_label = _("Deriver")
    _plugin_description = _(
        "Create events such as births and deaths by deriving their details from existing information."
    )

    @override
    def register_event_handlers(self, registry: EventHandlerRegistry) -> None:
        registry.add_handler(PostLoadAncestryEvent, _derive_ancestry)

    @override
    @classmethod
    def comes_before(cls) -> set[PluginIdentifier[Extension]]:
        return {Privatizer}



"""
Data types to reference files on disk.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

from typing_extensions import override

from betty.ancestry.file import File
from betty.locale.localizable import _, Localizable, ngettext
from betty.model import Entity
from betty.model.association import BidirectionalToOne, ToOneAssociate
from betty.plugin import ShorthandPluginBase

if TYPE_CHECKING:
    from betty.image import FocusArea
    from betty.ancestry.has_file_references import HasFileReferences


class FileReference(ShorthandPluginBase, Entity):
    """
    A reference between :py:class:`betty.ancestry.has_file_references.HasFileReferences` and betty.ancestry.file.File.

    This reference holds additional information specific to the relationship between the two entities.
    """

    _plugin_id = "file-reference"
    _plugin_label = _("File reference")

    #: The entity that references the file.
    referee = BidirectionalToOne["FileReference", "HasFileReferences"](
        "betty.ancestry.file_reference:FileReference",
        "referee",
        "betty.ancestry.has_file_references:HasFileReferences",
        "file_references",
        title="Referee",
        description="The entity referencing the file",
    )
    #: The referenced file.
    file = BidirectionalToOne["FileReference", File](
        "betty.ancestry.file_reference:FileReference",
        "file",
        "betty.ancestry.file:File",
        "referees",
        title="File",
        description="The file being referenced",
    )

    def __init__(
        self,
        referee: ToOneAssociate[HasFileReferences & Entity],
        file: ToOneAssociate[File],
        *,
        focus: FocusArea | None = None,
    ):
        super().__init__()
        self.referee = referee
        self.file = file
        self.focus = focus

    @override
    @classmethod
    def plugin_label_plural(cls) -> Localizable:
        return _("File references")

    @override
    @classmethod
    def plugin_label_count(cls, count: int) -> Localizable:
        return ngettext(
            "{count} file reference", "{count} file references", count
        ).format(count=str(count))

    @property
    def focus(self) -> FocusArea | None:
        """
        The area within the 2-dimensional representation of the file to focus on.

        This can be used to locate where faces are in a photo, or a specific article in a newspaper scan, for example.
        """
        return self._focus

    @focus.setter
    def focus(self, focus: FocusArea | None) -> None:
        self._focus = focus



"""
Provide an API to determine if information should be kept private.
"""

from __future__ import annotations

import logging
from contextlib import suppress
from datetime import datetime
from typing import Iterator, TypeAlias, Any, TYPE_CHECKING

from betty.ancestry.presence import Presence
from betty.ancestry.event import Event
from betty.ancestry.event_type.event_types import EndOfLifeEventType
from betty.ancestry.has_citations import HasCitations
from betty.ancestry.has_file_references import HasFileReferences
from betty.ancestry.has_notes import HasNotes
from betty.ancestry.person import Person
from betty.ancestry.place import Place
from betty.ancestry.presence_role.presence_roles import Subject
from betty.privacy import Privacy, HasPrivacy
from betty.ancestry.source import Source
from betty.date import DateRange, Date
from betty.model import Entity

if TYPE_CHECKING:
    from collections.abc import MutableSequence
    from betty.locale.localizer import Localizer

_Expirable: TypeAlias = Person | Event | Date | None


class Privatizer:
    """
    Privatize resources.
    """

    def __init__(
        self,
        lifetime_threshold: int,
        *,
        localizer: Localizer,
    ):
        self._lifetime_threshold = lifetime_threshold
        self._localizer = localizer
        self._seen: MutableSequence[HasPrivacy] = []

    def privatize(self, subject: HasPrivacy) -> None:
        """
        Privatize a resource.
        """
        if subject.privacy is Privacy.PUBLIC:
            return

        if isinstance(subject, Person):
            self._determine_person_privacy(subject)

        if isinstance(subject, Place):
            self._determine_place_privacy(subject)

        if subject.privacy is not Privacy.PRIVATE:
            return

        if subject in self._seen:
            return
        self._seen.append(subject)

        if isinstance(subject, Person):
            self._privatize_person(subject)

        if isinstance(subject, Presence):
            self._privatize_presence(subject)

        if isinstance(subject, Event):
            self._privatize_event(subject)

        if isinstance(subject, Place):
            self._privatize_place(subject)

        if isinstance(subject, Source):
            self._privatize_source(subject)

        if isinstance(subject, HasCitations):
            self._privatize_has_citations(subject)

        if isinstance(subject, HasFileReferences):
            self._privatize_has_file_references(subject)

        if isinstance(subject, HasNotes):
            self._privatize_has_notes(subject)

    def _privatize_person(self, person: Person) -> None:
        if not person.private:
            return

        for person_name in person.names:
            self._mark_private(person_name, person)
            self.privatize(person_name)
        for presence in person.presences:
            self._mark_private(presence, person)
            self.privatize(presence)

    def _privatize_presence(self, presence: Presence) -> None:
        if not presence.private:
            return

        if isinstance(presence.role, Subject):
            self._mark_private(presence.event, presence)
            self.privatize(presence.event)
        self._mark_private(presence.person, presence)
        self.privatize(presence.person)

    def _privatize_event(self, event: Event) -> None:
        if not event.private:
            return

        for presence in event.presences:
            self._mark_private(presence, event)
            self.privatize(presence)
        if event.place:
            self.privatize(event.place)

    def _privatize_place(self, place: Place) -> None:
        if not place.private:
            return

        for enclosure in place.enclosees:
            self._mark_private(enclosure.enclosee, place)
            self.privatize(enclosure.enclosee)
        for enclosure in place.enclosers:
            self.privatize(enclosure.encloser)

    def _privatize_has_citations(
        self, has_citations: HasCitations & HasPrivacy
    ) -> None:
        if not has_citations.private:
            return

        for citation in has_citations.citations:
            self._mark_private(citation, has_citations)
            self.privatize(citation)

    def _privatize_source(self, source: Source) -> None:
        if not source.private:
            return

        for contained_source in source.contains:
            self._mark_private(contained_source, source)
            self.privatize(contained_source)
        for citation in source.citations:
            self._mark_private(citation, source)
            self.privatize(citation)

    def _privatize_has_file_references(
        self, has_file_references: HasFileReferences & HasPrivacy
    ) -> None:
        if not has_file_references.private:
            return

        for file_reference in has_file_references.file_references:
            self._mark_private(file_reference.file, has_file_references)
            self.privatize(file_reference.file)

    def _privatize_has_notes(self, has_notes: HasNotes & HasPrivacy) -> None:
        if not has_notes.private:
            return

        for note in has_notes.notes:
            self._mark_private(note, has_notes)
            self.privatize(note)

    def _ancestors_by_generation(
        self, person: Person, generations_ago: int = 1
    ) -> Iterator[tuple[Person, int]]:
        for parent in person.parents:
            yield parent, generations_ago
            yield from self._ancestors_by_generation(parent, generations_ago + 1)

    def _determine_person_privacy(self, person: Person) -> None:
        # Do not change existing explicit privacy declarations.
        if person.privacy is not Privacy.UNDETERMINED:
            return

        # A dead person is not private, regardless of when they died.
        for presence in person.presences:
            if isinstance(presence.event.event_type, EndOfLifeEventType):
                if presence.event.date is None:
                    person.public = True
                    return
                if self.has_expired(presence.event, 0):
                    person.public = True
                    return

        if self.has_expired(person, 1):
            person.public = True
            return

        for ancestor, generations_ago in self._ancestors_by_generation(person):
            if self.has_expired(ancestor, generations_ago + 1):
                person.public = True
                return

        # If any descendant has any expired event, the person is considered not private.
        for descendant in person.descendants:
            if self.has_expired(descendant, 1):
                person.public = True
                return

        person.private = True
        logging.getLogger(__name__).debug(
            self._localizer._(
                "Privatized person {privatized_person_id} ({privatized_person}) because they are likely still alive."
            ).format(
                privatized_person_id=person.id,
                privatized_person=person.label.localize(self._localizer),
            )
        )

    def _determine_place_privacy(self, place: Place) -> None:
        # Do not change existing explicit privacy declarations.
        if place.privacy is not Privacy.UNDETERMINED:
            return

        # If there are non-private events, we will not privatize the place.
        for event in place.events:
            if not event.private:
                return

        # If there are non-private enclosed places, we will not privatize the place.
        for enclosure in place.enclosees:
            if not enclosure.enclosee.private:
                return

        place.private = True
        logging.getLogger(__name__).debug(
            self._localizer._(
                "Privatized place {privatized_place_id} ({privatized_place}) because it is not associated with any public information."
            ).format(
                privatized_place_id=place.id,
                privatized_place=place.label.localize(self._localizer),
            )
        )

    def has_expired(
        self,
        subject: _Expirable,
        generations_ago: int = 0,
    ) -> bool:
        """
        Check if a subject of the given generation has expired.
        """
        if isinstance(subject, Person):
            return self._person_has_expired(subject, generations_ago)

        if isinstance(subject, Event):
            return self._event_has_expired(subject, generations_ago)

        if isinstance(subject, Date):
            return self._date_has_expired(subject, generations_ago)

        return False

    def _person_has_expired(self, person: Person, generations_ago: int) -> bool:
        for presence in person.presences:
            if self._event_has_expired(presence.event, generations_ago):
                return True
        return False

    def _event_has_expired(self, event: Event, generations_ago: int) -> bool:
        date = event.date

        if isinstance(date, DateRange):
            # We can only determine event expiration with certainty if we have an end date to work with. Someone born in
            # 2000 can have a valid birth event with a start date of 1800, which does nothing to help us determine
            # expiration.
            date = date.end

        return self.has_expired(date, generations_ago)

    def _date_has_expired(
        self,
        date: Date,
        generations_ago: int,
    ) -> bool:
        if not date.comparable:
            return False

        return date <= Date(
            datetime.now().year - self._lifetime_threshold * generations_ago,
            datetime.now().month,
            datetime.now().day,
        )

    def _mark_private(self, target: HasPrivacy, reason: Any) -> None:
        # Do not change existing explicit privacy declarations.
        if target.own_privacy is not Privacy.UNDETERMINED:
            return

        target.private = True
        with suppress(ValueError):
            self._seen.remove(target)

        if isinstance(target, Entity) and isinstance(reason, Entity):
            logging.getLogger(__name__).debug(
                self._localizer._(
                    "Privatized {privatized_entity_type} {privatized_entity_id} ({privatized_entity}) because of {reason_entity_type} {reason_entity_id} ({reason_entity})."
                ).format(
                    privatized_entity_type=target.plugin_label().localize(
                        self._localizer
                    ),
                    privatized_entity_id=target.id,
                    privatized_entity=target.label.localize(self._localizer),
                    reason_entity_type=reason.plugin_label().localize(self._localizer),
                    reason_entity_id=reason.id,
                    reason_entity=reason.label.localize(self._localizer),
                )
            )



from __future__ import annotations  # noqa D100

import asyncio
from pathlib import Path
from typing import TYPE_CHECKING, final, Self

import asyncclick as click
from betty.app.factory import AppDependentFactory
from betty.cli.commands import command, Command
from betty.locale.localizable import _
from betty.plugin import ShorthandPluginBase
import betty.project.extension.demo as stddemo
from betty.project.extension.demo.project import create_project
from typing_extensions import override

if TYPE_CHECKING:
    from betty.app import App


@final
class Demo(ShorthandPluginBase, AppDependentFactory, Command):
    """
    A command to run the demonstration site.
    """

    _plugin_id = "demo"
    _plugin_label = _("Explore a demonstration site")

    def __init__(self, app: App):
        self._app = app

    @override
    @classmethod
    async def new_for_app(cls, app: App) -> Self:
        return cls(app)

    @override
    async def click_command(self) -> click.Command:
        localizer = await self._app.localizer
        description = self.plugin_description()

        @command(
            self.plugin_id(),
            short_help=self.plugin_label().localize(localizer),
            help=description.localize(localizer)
            if description
            else self.plugin_label().localize(localizer),
        )
        @click.option(
            "--path",
            "path",
            help="The path to the project directory to generate the demonstration site into instead of serving the site in a browser window.",
        )
        @click.option(
            "--url",
            "url",
            help="The site's public project URL. Used only when `--path` is given.",
        )
        async def demo(*, path: str | None, url: str | None) -> None:
            from betty.project.extension.demo.serve import DemoServer

            if path is None:
                async with DemoServer(app=self._app) as server:
                    await server.show()
                    while True:
                        await asyncio.sleep(999)
            else:
                project = await create_project(self._app, Path(path))
                if url is not None:
                    project.configuration.url = url
                async with project:
                    await stddemo.generate_with_cleanup(project)

        return demo



"""
Data types describing persons.
"""

from __future__ import annotations

from typing import final, Iterable, MutableSequence, Iterator, TYPE_CHECKING
from urllib.parse import quote

from typing_extensions import override

from betty.ancestry.citation import Citation
from betty.ancestry.gender.genders import Unknown as UnknownGender
from betty.ancestry.has_citations import HasCitations
from betty.ancestry.has_file_references import HasFileReferences
from betty.ancestry.has_notes import HasNotes
from betty.ancestry.link import HasLinks, Link
from betty.functools import unique
from betty.json.linked_data import dump_context, JsonLdObject
from betty.locale.localizable import _, Localizable, ngettext
from betty.model import (
    UserFacingEntity,
    Entity,
    persistent_id,
    ToManySchema,
)
from betty.model.association import BidirectionalToMany, ToManyAssociates
from betty.plugin import ShorthandPluginBase
from betty.privacy import HasPrivacy, Privacy

if TYPE_CHECKING:
    from betty.mutability import Mutable
    from betty.ancestry.person_name import PersonName
    from betty.ancestry.note import Note
    from betty.ancestry.file_reference import FileReference
    from betty.ancestry.citation import Citation
    from betty.ancestry.presence import Presence
    from betty.ancestry.gender import Gender
    from betty.project import Project
    from betty.serde.dump import DumpMapping, Dump


@final
class Person(
    ShorthandPluginBase,
    HasFileReferences,
    HasCitations,
    HasNotes,
    HasLinks,
    HasPrivacy,
    UserFacingEntity,
    Entity,
):
    """
    A person.
    """

    _plugin_id = "person"
    _plugin_label = _("Person")

    parents = BidirectionalToMany["Person", "Person"](
        "betty.ancestry.person:Person",
        "parents",
        "betty.ancestry.person:Person",
        "children",
        title="Parents",
    )
    children = BidirectionalToMany["Person", "Person"](
        "betty.ancestry.person:Person",
        "children",
        "betty.ancestry.person:Person",
        "parents",
        title="Children",
    )
    presences = BidirectionalToMany["Person", "Presence"](
        "betty.ancestry.person:Person",
        "presences",
        "betty.ancestry.presence:Presence",
        "person",
        title="Presences",
        description="This person's presences at events",
        linked_data_embedded=True,
    )
    names = BidirectionalToMany["Person", "PersonName"](
        "betty.ancestry.person:Person",
        "names",
        "betty.ancestry.person_name:PersonName",
        "person",
        title="Names",
        linked_data_embedded=True,
    )

    def __init__(
        self,
        *,
        id: str | None = None,  # noqa A002
        file_references: ToManyAssociates[FileReference] | None = None,
        citations: ToManyAssociates["Citation"] | None = None,
        links: MutableSequence[Link] | None = None,
        notes: ToManyAssociates[Note] | None = None,
        privacy: Privacy | None = None,
        public: bool | None = None,
        private: bool | None = None,
        parents: ToManyAssociates[Person] | None = None,
        children: ToManyAssociates[Person] | None = None,
        presences: ToManyAssociates["Presence"] | None = None,
        names: ToManyAssociates["PersonName"] | None = None,
        gender: Gender | None = None,
    ):
        super().__init__(
            id,
            file_references=file_references,
            citations=citations,
            links=links,
            notes=notes,
            privacy=privacy,
            public=public,
            private=private,
        )
        if children is not None:
            self.children = children
        if parents is not None:
            self.parents = parents
        if presences is not None:
            self.presences = presences
        if names is not None:
            self.names = names
        self._gender = gender or UnknownGender()

    @override
    @classmethod
    def plugin_label_plural(cls) -> Localizable:
        return _("People")

    @override
    @classmethod
    def plugin_label_count(cls, count: int) -> Localizable:
        return ngettext("{count} person", "{count} people", count).format(
            count=str(count)
        )

    @override
    def get_mutable_instances(self) -> Iterable[Mutable]:
        return (
            *super().get_mutable_instances(),
            self.gender,
        )

    @property
    def gender(self) -> Gender:
        """
        The person's gender.
        """
        return self._gender

    @gender.setter
    def gender(self, gender: Gender) -> None:
        self.assert_mutable()
        self._gender = gender

    @property
    def ancestors(self) -> Iterator[Person]:
        """
        All ancestors.
        """
        for parent in self.parents:
            yield parent
            yield from parent.ancestors

    @property
    def siblings(self) -> Iterator[Person]:
        """
        All siblings.
        """
        yield from unique(
            sibling
            for parent in self.parents
            for sibling in parent.children
            if sibling != self
        )

    @property
    def descendants(self) -> Iterator[Person]:
        """
        All descendants.
        """
        for child in self.children:
            yield child
            yield from child.descendants

    @override
    @property
    def label(self) -> Localizable:
        for name in self.names:
            if name.public:
                return name.label
        return super().label

    @override
    async def dump_linked_data(self, project: Project) -> DumpMapping[Dump]:
        dump = await super().dump_linked_data(project)
        url_generator = await project.url_generator
        dump_context(
            dump,
            names="https://schema.org/name",
            parents="https://schema.org/parent",
            children="https://schema.org/child",
            siblings="https://schema.org/sibling",
        )
        dump["@type"] = "https://schema.org/Person"
        dump["siblings"] = [
            url_generator.generate(
                f"betty-static:///person/{quote(sibling.id)}/index.json"
            )
            for sibling in self.siblings
            if persistent_id(sibling)
        ]
        if self.public:
            dump["gender"] = self.gender.plugin_id()
        return dump

    @override
    @classmethod
    async def linked_data_schema(cls, project: Project) -> JsonLdObject:
        schema = await super().linked_data_schema(project)
        schema.add_property(
            "gender", await project.gender_repository.plugin_id_schema, False
        )
        schema.add_property("siblings", ToManySchema(title="Siblings"))
        return schema



from abc import abstractmethod
from collections.abc import Sequence, AsyncIterator
from contextlib import asynccontextmanager
from datetime import datetime
from multiprocessing.managers import SyncManager
from typing import (
    Generic,
    Self,
    overload,
    AsyncContextManager,
    Literal,
    TypeVar,
    Protocol,
    Any,
)

from typing_extensions import override

from betty.cache import Cache, CacheItem, CacheItemValueSetter
from betty.concurrent import AsynchronizedLock, Ledger
from betty.concurrent import ensure_manager
from betty.typing import processsafe

_CacheT = TypeVar("_CacheT", bound=Cache[Any])
_CacheItemValueCoT = TypeVar("_CacheItemValueCoT", covariant=True)
_CacheItemValueContraT = TypeVar("_CacheItemValueContraT", contravariant=True)


class _CommonCacheBaseGetter(Generic[_CacheItemValueCoT], Protocol):
    async def __call__(
        self, cache_item_id: str
    ) -> CacheItem[_CacheItemValueCoT] | None:
        pass


class _CommonCacheBaseSetter(Generic[_CacheItemValueContraT], Protocol):
    async def __call__(
        self,
        cache_item_id: str,
        value: _CacheItemValueContraT,
        *,
        modified: int | float | None = None,
    ) -> None:
        pass


class _StaticCacheItem(CacheItem[_CacheItemValueCoT], Generic[_CacheItemValueCoT]):
    __slots__ = "_value", "_modified"

    def __init__(
        self,
        value: _CacheItemValueCoT,
        modified: int | float | None = None,
    ):
        self._value = value
        self._modified = datetime.now().timestamp() if modified is None else modified

    @override
    async def value(self) -> _CacheItemValueCoT:
        return self._value

    @override
    @property
    def modified(self) -> int | float:
        return self._modified


class _CommonCacheBaseState(Generic[_CacheT]):
    def __init__(
        self,
        cache_lock: AsynchronizedLock,
        cache_item_lock_ledger: Ledger,
    ):
        self.cache_lock = cache_lock
        self.cache_item_lock_ledger = cache_item_lock_ledger


@processsafe
class _CommonCacheBase(Cache[_CacheItemValueContraT], Generic[_CacheItemValueContraT]):
    def __init__(
        self,
        *,
        scopes: Sequence[str] | None = None,
        manager: SyncManager | _CommonCacheBaseState[Self] | None = None,
    ):
        self._scopes = scopes or ()
        if isinstance(manager, _CommonCacheBaseState):
            self._cache_lock = manager.cache_lock
            self._cache_item_lock_ledger = manager.cache_item_lock_ledger
        else:
            manager = ensure_manager(manager)
            self._cache_lock = AsynchronizedLock(manager.Lock())
            self._cache_item_lock_ledger = Ledger(self._cache_lock, manager=manager)

    @override
    @asynccontextmanager
    async def get(
        self, cache_item_id: str
    ) -> AsyncIterator[CacheItem[_CacheItemValueContraT] | None]:
        async with self._cache_item_lock_ledger.ledger(cache_item_id):
            yield await self._get(cache_item_id)

    @abstractmethod
    async def _get(
        self, cache_item_id: str
    ) -> CacheItem[_CacheItemValueContraT] | None:
        pass

    @override
    async def set(
        self,
        cache_item_id: str,
        value: _CacheItemValueContraT,
        *,
        modified: int | float | None = None,
    ) -> None:
        async with self._cache_item_lock_ledger.ledger(cache_item_id):
            await self._set(cache_item_id, value, modified=modified)

    @abstractmethod
    async def _set(
        self,
        cache_item_id: str,
        value: _CacheItemValueContraT,
        *,
        modified: int | float | None = None,
    ) -> None:
        pass

    @overload
    def getset(
        self, cache_item_id: str
    ) -> AsyncContextManager[
        tuple[
            CacheItem[_CacheItemValueContraT] | None,
            CacheItemValueSetter[_CacheItemValueContraT],
        ]
    ]:
        pass

    @overload
    def getset(
        self, cache_item_id: str, *, wait: Literal[False] = False
    ) -> AsyncContextManager[
        tuple[
            CacheItem[_CacheItemValueContraT] | None,
            CacheItemValueSetter[_CacheItemValueContraT] | None,
        ]
    ]:
        pass

    def getset(
        self, cache_item_id: str, *, wait: bool = True
    ) -> AsyncContextManager[
        tuple[
            CacheItem[_CacheItemValueContraT] | None,
            CacheItemValueSetter[_CacheItemValueContraT] | None,
        ]
    ]:
        return self._getset(cache_item_id, self._get, self._set, wait=wait)

    @asynccontextmanager
    async def _getset(
        self,
        cache_item_id: str,
        getter: _CommonCacheBaseGetter[_CacheItemValueContraT],
        setter: _CommonCacheBaseSetter[_CacheItemValueContraT],
        *,
        wait: bool = True,
    ) -> AsyncIterator[
        tuple[
            CacheItem[_CacheItemValueContraT] | None,
            CacheItemValueSetter[_CacheItemValueContraT] | None,
        ]
    ]:
        lock = self._cache_item_lock_ledger.ledger(cache_item_id)
        if await lock.acquire(wait=wait):
            try:

                async def _setter(value: _CacheItemValueContraT) -> None:
                    await setter(cache_item_id, value)

                yield await getter(cache_item_id), _setter
                return
            finally:
                await lock.release()
        yield None, None

    @override
    async def delete(self, cache_item_id: str) -> None:
        async with self._cache_item_lock_ledger.ledger(cache_item_id):
            await self._delete(cache_item_id)

    @abstractmethod
    async def _delete(self, cache_item_id: str) -> None:
        pass

    @override
    async def clear(self) -> None:
        async with self._cache_lock:
            await self._clear()

    @abstractmethod
    async def _clear(self) -> None:
        pass



"""
URL generators for project resources.
"""

from __future__ import annotations

from typing import final, Any, Self, TYPE_CHECKING
from urllib.parse import quote, urlparse

from typing_extensions import override

from betty.media_type.media_types import HTML, JSON, JSON_LD
from betty.project.factory import ProjectDependentFactory
from betty.string import camel_case_to_kebab_case
from betty.typing import private
from betty.url import (
    generate_from_path,
    LocalizedUrlGenerator as StdLocalizedUrlGenerator,
    StaticUrlGenerator as StdStaticUrlGenerator,
    PassthroughUrlGenerator,
    InvalidMediaType,
    UrlGenerator,
)
from betty.url.proxy import ProxyLocalizedUrlGenerator, ProxyUrlGenerator
from betty.model import Entity
from betty.warnings import deprecated

if TYPE_CHECKING:
    from betty.ancestry import Ancestry
    from betty.media_type import MediaType
    from betty.project import Project
    from betty.locale import Localey
    from collections.abc import Mapping


class _ProjectUrlGenerator(ProjectDependentFactory):
    def __init__(
        self,
        base_url: str,
        root_path: str,
        locales: Mapping[str, str],
        clean_urls: bool,
    ):
        self._base_url = base_url
        self._root_path = root_path
        self._locales = locales
        assert len(locales)
        self._default_locale = next(iter(locales))
        self._clean_urls = clean_urls

    @classmethod
    async def new_for_project(cls, project: Project) -> Self:
        """
        Create a new instance using the given project.
        """
        return cls(
            project.configuration.base_url,
            project.configuration.root_path,
            {
                locale_configuration.locale: locale_configuration.alias
                for locale_configuration in project.configuration.locales.values()
            },
            project.configuration.clean_urls,
        )

    def _generate_from_path(
        self, path: str, *, absolute: bool = False, locale: Localey | None = None
    ) -> str:
        return generate_from_path(
            path,
            absolute=absolute,
            locale=locale,
            base_url=self._base_url,
            root_path=self._root_path,
            locales=self._locales,
            clean_urls=self._clean_urls,
        )

    def _generate_from_entity(
        self,
        entity: Entity,
        pattern: str,
        *,
        media_type: MediaType | None,
        locale: Localey | None,
        absolute: bool,
    ) -> str:
        if media_type not in [HTML, JSON_LD, JSON]:
            raise InvalidMediaType.new(entity, media_type)
        extension, locale = _get_extension_and_locale(
            media_type, self._default_locale, locale=locale
        )
        return self._generate_from_path(
            pattern.format(
                entity_type=camel_case_to_kebab_case(entity.plugin_id()),
                entity_id=quote(entity.id),
                extension=extension,
            ),
            absolute=absolute,
            locale=locale,
        )

    def _generate_from_entity_type(
        self,
        entity_type: type[Entity],
        pattern: str,
        *,
        media_type: MediaType | None,
        locale: Localey | None,
        absolute: bool,
    ) -> str:
        if media_type not in [HTML, JSON_LD, JSON]:
            raise InvalidMediaType.new(entity_type, media_type)
        extension, locale = _get_extension_and_locale(
            media_type, self._default_locale, locale=locale
        )
        return self._generate_from_path(
            pattern.format(
                entity_type=camel_case_to_kebab_case(entity_type.plugin_id()),
                extension=extension,
            ),
            absolute=absolute,
            locale=locale,
        )


def _supports_path(resource: Any) -> bool:
    return isinstance(resource, str) and resource.startswith("/")


@final
class _LocalizedPathUrlGenerator(_ProjectUrlGenerator, StdLocalizedUrlGenerator):
    @override
    def supports(self, resource: Any) -> bool:
        return _supports_path(resource)

    @override
    def generate(
        self,
        resource: Any,
        media_type: MediaType,
        *,
        absolute: bool = False,
        locale: Localey | None = None,
    ) -> str:
        assert self.supports(resource)
        return self._generate_from_path(
            resource,
            absolute=absolute,
            locale=locale or self._default_locale,
        )


async def new_project_url_generator(project: Project) -> UrlGenerator:
    """
    Generate URLs for all resources provided by a Betty project.
    """
    entity_url_generator = await _EntityUrlGenerator.new_for_project(project)
    return ProxyUrlGenerator(
        await _EntityTypeUrlGenerator.new_for_project(project),
        entity_url_generator,
        _EntityUrlUrlGenerator(project.ancestry, entity_url_generator),
        await _LocalizedPathUrlUrlGenerator.new_for_project(project),
        await _StaticPathUrlUrlGenerator.new_for_project(project),
        PassthroughUrlGenerator(),
    )


@deprecated(
    f"This class has been deprecated since Betty 0.4.8, and will be removed in Betty 0.5. Instead use {new_project_url_generator}."
)
@final
class StaticUrlGenerator(_ProjectUrlGenerator, StdStaticUrlGenerator):
    """
    Generate URLs for static (non-localized) file paths.
    """

    @override
    def supports(self, resource: Any) -> bool:
        return _supports_path(resource)

    @override
    def generate(
        self,
        resource: Any,
        *,
        absolute: bool = False,
    ) -> str:
        assert self.supports(resource)
        return self._generate_from_path(resource, absolute=absolute)


def _get_extension_and_locale(
    media_type: MediaType, default_locale: str, *, locale: Localey | None
) -> tuple[str, Localey | None]:
    if media_type == HTML:
        return "html", locale or default_locale
    elif media_type in (JSON, JSON_LD):
        return "json", None
    else:
        raise ValueError(f'Unknown entity media type "{media_type}".')


class __EntityTypeUrlGenerator(_ProjectUrlGenerator):
    _pattern = "/{entity_type}/index.{extension}"

    def supports(self, resource: Any) -> bool:
        return isinstance(resource, type) and issubclass(resource, Entity)


@final
class _EntityTypeUrlGenerator(__EntityTypeUrlGenerator, UrlGenerator):
    @override
    def generate(
        self,
        resource: type[Entity],
        *,
        media_type: MediaType | None = None,
        absolute: bool = False,
        locale: Localey | None = None,
    ) -> str:
        assert self.supports(resource)
        return self._generate_from_entity_type(
            resource,
            self._pattern,
            media_type=media_type,
            locale=locale,
            absolute=absolute,
        )


@final
class _EntityTypeLocalizedUrlGenerator(
    __EntityTypeUrlGenerator, StdLocalizedUrlGenerator
):
    @override
    def generate(
        self,
        resource: type[Entity],
        media_type: MediaType,
        *,
        absolute: bool = False,
        locale: Localey | None = None,
    ) -> str:
        assert self.supports(resource)
        return self._generate_from_entity_type(
            resource,
            self._pattern,
            media_type=media_type,
            locale=locale,
            absolute=absolute,
        )


class __EntityUrlGenerator(_ProjectUrlGenerator):
    _pattern = "/{entity_type}/{entity_id}/index.{extension}"

    def supports(self, resource: Any) -> bool:
        return isinstance(resource, Entity)


@final
class _EntityUrlGenerator(__EntityUrlGenerator, UrlGenerator):
    @override
    def generate(
        self,
        resource: Entity,
        *,
        media_type: MediaType | None = None,
        absolute: bool = False,
        locale: Localey | None = None,
    ) -> str:
        assert self.supports(resource)
        return self._generate_from_entity(
            resource,
            self._pattern,
            media_type=media_type,
            locale=locale,
            absolute=absolute,
        )


@final
class _EntityLocalizedUrlGenerator(__EntityUrlGenerator, StdLocalizedUrlGenerator):
    @override
    def generate(
        self,
        resource: Entity,
        media_type: MediaType,
        *,
        absolute: bool = False,
        locale: Localey | None = None,
    ) -> str:
        assert self.supports(resource)
        return self._generate_from_entity(
            resource,
            self._pattern,
            media_type=media_type,
            locale=locale,
            absolute=absolute,
        )


class _EntityUrlUrlGenerator(UrlGenerator):
    def __init__(self, ancestry: Ancestry, entity_url_generator: _EntityUrlGenerator):
        self._ancestry = ancestry
        self._entity_url_generator = entity_url_generator

    @override
    def supports(self, resource: Any) -> bool:
        if not isinstance(resource, str):
            return False
        try:
            parsed_url = urlparse(resource)
        except ValueError:
            return False
        if parsed_url.scheme != "betty-entity":
            return False
        if not parsed_url.netloc:
            return False
        if not len(parsed_url.path) >= 2:
            return False
        return True

    @override
    def generate(
        self,
        resource: str,
        *,
        media_type: MediaType | None = None,
        absolute: bool = False,
        locale: Localey | None = None,
    ) -> str:
        parsed_url = urlparse(resource)
        entity_type_id = parsed_url.netloc
        entity_id = parsed_url.path[1:]
        entity = self._ancestry[entity_type_id][entity_id]
        return self._entity_url_generator.generate(
            entity, media_type=media_type, absolute=absolute, locale=locale
        )


class _LocalizedPathUrlUrlGenerator(_ProjectUrlGenerator, UrlGenerator):
    @override
    def supports(self, resource: Any) -> bool:
        if not isinstance(resource, str):
            return False
        try:
            parsed_url = urlparse(resource)
        except ValueError:
            return False
        if parsed_url.scheme != "betty":
            return False
        if not parsed_url.netloc and not parsed_url.path:
            return False
        return True

    @override
    def generate(
        self,
        resource: str,
        *,
        media_type: MediaType | None = None,
        absolute: bool = False,
        locale: Localey | None = None,
    ) -> str:
        assert self.supports(resource)
        parsed_url = urlparse(resource)
        url_path = "/" + (parsed_url.netloc + parsed_url.path).lstrip("/")
        return self._generate_from_path(
            url_path,
            absolute=absolute,
            locale=locale or self._default_locale,
        )


class _StaticPathUrlUrlGenerator(_ProjectUrlGenerator, UrlGenerator):
    @override
    def supports(self, resource: Any) -> bool:
        if not isinstance(resource, str):
            return False
        try:
            parsed_url = urlparse(resource)
        except ValueError:
            return False
        if parsed_url.scheme != "betty-static":
            return False
        if not parsed_url.netloc and not parsed_url.path:
            return False
        return True

    @override
    def generate(
        self,
        resource: str,
        *,
        media_type: MediaType | None = None,
        absolute: bool = False,
        locale: Localey | None = None,
    ) -> str:
        assert self.supports(resource)
        parsed_url = urlparse(resource)
        url_path = "/" + (parsed_url.netloc + parsed_url.path).lstrip("/")
        return self._generate_from_path(url_path, absolute=absolute)


@deprecated(
    f"This class has been deprecated since Betty 0.4.8, and will be removed in Betty 0.5. Instead use {UrlGenerator}."
)
@final
class LocalizedUrlGenerator(StdLocalizedUrlGenerator, ProjectDependentFactory):
    """
    Generate URLs for all resources provided by a Betty project.
    """

    @private
    def __init__(
        self,
        *upstreams: StdLocalizedUrlGenerator,
    ):
        self._upstream = ProxyLocalizedUrlGenerator(*upstreams)

    @override
    @classmethod
    async def new_for_project(cls, project: Project) -> Self:
        return cls(
            await _EntityTypeLocalizedUrlGenerator.new_for_project(project),
            await _EntityLocalizedUrlGenerator.new_for_project(project),
            await _LocalizedPathUrlGenerator.new_for_project(project),
        )

    @override
    def supports(self, resource: Any) -> bool:
        return self._upstream.supports(resource)

    @override
    def generate(
        self,
        resource: Any,
        media_type: MediaType,
        *,
        absolute: bool = False,
        locale: Localey | None = None,
    ) -> str:
        return self._upstream.generate(
            resource, media_type, absolute=absolute, locale=locale
        )



"""
Provide Locale API errors.
"""

from __future__ import annotations

from babel.core import Locale
from babel.localedata import locale_identifiers

from betty.error import UserFacingError
from betty.locale import to_locale
from betty.locale.localizable import _, join, do_you_mean


class LocaleError(UserFacingError, Exception):
    """
    A locale API error.
    """

    pass


class InvalidLocale(LocaleError, ValueError):
    """
    Raised when a value is not a valid locale.
    """

    def __init__(self, invalid_locale: str) -> None:
        super().__init__(
            _('"{invalid_locale}" is not a valid IETF BCP 47 language tag.').format(
                invalid_locale=invalid_locale
            )
        )
        self.invalid_locale = invalid_locale


class LocaleNotFound(LocaleError, ValueError):
    """
    Raised when a locale could not be found.
    """

    def __init__(self, locale: str) -> None:
        locale_chars = {char for char in locale[: locale.find("-")] if char.isalpha()}
        available_locales = sorted(
            to_locale(Locale.parse(identifier))
            for identifier in locale_identifiers()
            if set(identifier[: identifier.find("_")]) & locale_chars
        )
        super().__init__(
            join(
                _("Cannot find locale {locale}.").format(locale=locale),
                do_you_mean(*available_locales),
            )
        )
        self.locale = locale



"""
Provide a URL generation API.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, TYPE_CHECKING, Self
from urllib.parse import urlparse

from typing_extensions import override, deprecated

from betty.locale import negotiate_locale, Localey, to_locale

if TYPE_CHECKING:
    from betty.media_type import MediaType
    from collections.abc import Mapping


class GenerationError(RuntimeError):
    """
    A URL generation error.
    """

    pass


class UnsupportedResource(GenerationError):
    """
    Raised when a URL generator cannot generate a URL for a resource.

    These are preventable by checking :py:meth:`betty.url.LocalizedUrlGenerator.supports` or
    :py:meth:`betty.url.StaticUrlGenerator.supports` first.
    """

    @classmethod
    def new(cls, resource: Any) -> Self:
        """
        Create a new instance.
        """
        return cls(f"Unsupported resource: {resource}")


class InvalidMediaType(GenerationError):
    """
    Raised when a URL generator cannot generate a URL for a resource with the given media type.
    """

    @classmethod
    def new(cls, resource: Any, media_type: MediaType | None) -> Self:
        """
        Create a new instance.
        """
        if media_type:
            return cls(f"Unsupported media type '{media_type}' for resource {resource}")
        return cls(f"Missing media type for resource {resource}")


class _UrlGenerator(ABC):
    """
    Generate URLs for localizable resources.
    """

    @abstractmethod
    def supports(self, resource: Any) -> bool:
        """
        Whether the given resource is supported by this URL generator.
        """
        pass


class UrlGenerator(_UrlGenerator):
    """
    Generate URLs for resources.
    """

    @abstractmethod
    def generate(
        self,
        resource: Any,
        *,
        media_type: MediaType | None = None,
        absolute: bool = False,
        locale: Localey | None = None,
    ) -> str:
        """
        Generate a URL for a resource.

        :raise UnsupportedResource:
        :raise InvalidMediaType:
        """
        pass


@deprecated(
    f"This class has been deprecated since Betty 0.4.8, and will be removed in Betty 0.5. Instead use {UrlGenerator}."
)
class LocalizedUrlGenerator(_UrlGenerator):
    """
    Generate URLs for localizable resources.
    """

    @abstractmethod
    def generate(
        self,
        resource: Any,
        media_type: MediaType,
        *,
        absolute: bool = False,
        locale: Localey | None = None,
    ) -> str:
        """
        Generate a URL for a resource.

        :raise UnsupportedResource:
        :raise InvalidMediaType:
        """
        pass


class PassthroughUrlGenerator(UrlGenerator):
    """
    Returns resources verbatim if they are absolute URLs already.
    """

    @override
    def supports(self, resource: Any) -> bool:
        if not isinstance(resource, str):
            return False
        try:
            return bool(urlparse(resource).scheme)
        except ValueError:
            return False

    @override
    def generate(
        self,
        resource: Any,
        *,
        media_type: MediaType | None = None,
        absolute: bool = False,
        locale: Localey | None = None,
    ) -> str:
        assert isinstance(resource, str)
        return resource


@deprecated(
    f"This class has been deprecated since Betty 0.4.8, and will be removed in Betty 0.5. Instead use {UrlGenerator}."
)
class StaticUrlGenerator(_UrlGenerator):
    """
    Generate URLs for static (non-localizable) resources.
    """

    @abstractmethod
    def generate(
        self,
        resource: Any,
        *,
        absolute: bool = False,
    ) -> str:
        """
        Generate a static URL for a static resource.

        :raise UnsupportedResource:
        :raise InvalidMediaType:
        """
        pass


def generate_from_path(
    path: str,
    *,
    base_url: str,
    root_path: str,
    locales: Mapping[str, str],
    clean_urls: bool,
    absolute: bool = False,
    locale: Localey | None = None,
) -> str:
    """
    Generate a full URL from a public path.
    """
    url = base_url.rstrip("/") if absolute else ""
    url += root_path.rstrip("/")
    assert path.startswith("/"), (
        f'Paths must be root-relative (start with a forward slash), but "{path}" was given'
    )
    path = path.strip("/")
    if locale and len(locales) > 1:
        locale = to_locale(locale)
        try:
            negotiated_locale_data = negotiate_locale(locale, list(locales))
            if negotiated_locale_data is None:
                raise KeyError
            locale_alias = locales[to_locale(negotiated_locale_data)]
        except KeyError:
            raise ValueError(
                f'Cannot generate URLs in "{locale}", because it cannot be resolved to any of the available locales: {", ".join(locales)}'
            ) from None
        url += f"/{locale_alias}"
    if path:
        url += f"/{path}"
    if clean_urls and url.endswith("/index.html"):
        url = url[:-11]
    # Ensure URLs are root-relative.
    if not absolute:
        url = f"/{url.lstrip('/')}"
    return url



"""
Provide the place entity.
"""

from __future__ import annotations

from contextlib import suppress
from typing import final, MutableSequence, Iterable, Iterator, TYPE_CHECKING

from typing_extensions import override

from betty.ancestry.has_file_references import HasFileReferences
from betty.ancestry.has_notes import HasNotes
from betty.ancestry.link import HasLinks, Link
from betty.ancestry.name import Name
from betty.ancestry.place_type.place_types import Unknown as UnknownPlaceType
from betty.json.linked_data import dump_context, JsonLdObject
from betty.json.schema import Array, Number, Object
from betty.locale.localizable import _, Localizable, ngettext
from betty.model import UserFacingEntity, Entity
from betty.model.association import BidirectionalToMany, ToManyAssociates
from betty.plugin import ShorthandPluginBase
from betty.privacy import HasPrivacy

if TYPE_CHECKING:
    from betty.mutability import Mutable
    from betty.ancestry.note import Note
    from betty.ancestry.event import Event
    from betty.ancestry.enclosure import Enclosure
    from betty.ancestry.place_type import PlaceType
    from betty.privacy import Privacy
    from geopy import Point
    from betty.project import Project
    from betty.serde.dump import DumpMapping, Dump


@final
class Place(
    ShorthandPluginBase,
    HasLinks,
    HasFileReferences,
    HasNotes,
    HasPrivacy,
    UserFacingEntity,
    Entity,
):
    """
    A place.

    A place is a physical location on earth. It may be identifiable by GPS coordinates only, or
    be a well-known city, with names in many languages, imagery, and its own Wikipedia page, or
    any type of place in between.
    """

    _plugin_id = "place"
    _plugin_label = _("Place")

    events = BidirectionalToMany["Place", "Event"](
        "betty.ancestry.place:Place",
        "events",
        "betty.ancestry.event:Event",
        "place",
        title="Events",
        description="The events that happened in this place",
    )
    enclosers = BidirectionalToMany["Place", "Enclosure"](
        "betty.ancestry.place:Place",
        "encloser",
        "betty.ancestry.enclosure:Enclosure",
        "enclosee",
        title="Enclosers",
        description="The places this place is enclosed or contained by",
        linked_data_embedded=True,
    )
    enclosees = BidirectionalToMany["Place", "Enclosure"](
        "betty.ancestry.place:Place",
        "enclosee",
        "betty.ancestry.enclosure:Enclosure",
        "encloser",
        title="Enclosees",
        description="The places this place encloses or contains",
        linked_data_embedded=True,
    )

    def __init__(
        self,
        *,
        id: str | None = None,  # noqa A002
        names: MutableSequence[Name] | None = None,
        events: ToManyAssociates[Event] | None = None,
        enclosers: ToManyAssociates["Enclosure"] | None = None,
        enclosees: ToManyAssociates["Enclosure"] | None = None,
        notes: ToManyAssociates[Note] | None = None,
        coordinates: Point | None = None,
        links: MutableSequence[Link] | None = None,
        privacy: Privacy | None = None,
        public: bool | None = None,
        private: bool | None = None,
        place_type: PlaceType | None = None,
    ):
        super().__init__(
            id,
            notes=notes,
            links=links,
            privacy=privacy,
            public=public,
            private=private,
        )
        self._names = [] if names is None else names
        self._coordinates = coordinates
        if events is not None:
            self.events = events
        if enclosers is not None:
            self.enclosers = enclosers
        if enclosees is not None:
            self.enclosees = enclosees
        self._place_type = place_type or UnknownPlaceType()

    @override
    def get_mutable_instances(self) -> Iterable[Mutable]:
        return (
            *super().get_mutable_instances(),
            self._place_type,
        )

    @property
    def walk_enclosees(self) -> Iterator["Enclosure"]:
        """
        All enclosed places.
        """
        for enclosure in self.enclosees:
            yield enclosure
            yield from enclosure.enclosee.walk_enclosees

    @override
    @classmethod
    def plugin_label_plural(cls) -> Localizable:
        return _("Places")

    @override
    @classmethod
    def plugin_label_count(cls, count: int) -> Localizable:
        return ngettext("{count} place", "{count} places", count).format(
            count=str(count)
        )

    @property
    def place_type(self) -> PlaceType:
        """
        The type of this place.
        """
        return self._place_type

    @place_type.setter
    def place_type(self, place_type: PlaceType) -> None:
        self.assert_mutable()
        self._place_type = place_type

    @property
    def names(self) -> MutableSequence[Name]:
        """
        The place's names.

        The first name is considered the :py:attr:`place label <betty.ancestry.place.Place.label>`.
        """
        return self._names

    @property
    def coordinates(self) -> Point | None:
        """
        The place's coordinates.
        """
        return self._coordinates

    @coordinates.setter
    def coordinates(self, coordinates: Point):
        self.assert_mutable()
        self._coordinates = coordinates

    @override
    @property
    def label(self) -> Localizable:
        with suppress(IndexError):
            return self.names[0].name
        return super().label

    @override
    async def dump_linked_data(self, project: Project) -> DumpMapping[Dump]:
        dump = await super().dump_linked_data(project)
        dump_context(
            dump,
            names="https://schema.org/name",
            events="https://schema.org/event",
            enclosers="https://schema.org/containedInPlace",
            enclosees="https://schema.org/containsPlace",
        )
        dump["@type"] = "https://schema.org/Place"
        dump["names"] = [await name.dump_linked_data(project) for name in self.names]
        if self.coordinates is not None:
            dump["coordinates"] = {
                "@type": "https://schema.org/GeoCoordinates",
                "latitude": self.coordinates.latitude,
                "longitude": self.coordinates.longitude,
            }
            dump_context(dump, coordinates="https://schema.org/geo")
            dump_context(
                dump["coordinates"],  # type: ignore[arg-type]
                latitude="https://schema.org/latitude",
            )
            dump_context(
                dump["coordinates"],  # type: ignore[arg-type]
                longitude="https://schema.org/longitude",
            )
        return dump

    @override
    @classmethod
    async def linked_data_schema(cls, project: Project) -> JsonLdObject:
        schema = await super().linked_data_schema(project)
        schema.add_property(
            "names", Array(await Name.linked_data_schema(project), title="Names")
        )
        coordinate_schema = Number(title="Coordinate")
        coordinates_schema = Object(title="Coordinates")
        coordinates_schema.add_property("latitude", coordinate_schema, False)
        coordinates_schema.add_property("longitude", coordinate_schema, False)
        schema.add_property("coordinates", coordinates_schema, False)
        return schema



"""
An API for providing application-wide services.
"""

from abc import ABC, abstractmethod
from collections.abc import Callable, MutableSequence, Awaitable
from inspect import iscoroutinefunction, getmembers
from types import TracebackType
from typing import (
    Self,
    Any,
    final,
    TypedDict,
    Unpack,
    TypeAlias,
    cast,
    Generic,
    Protocol,
)
from typing import overload, TypeVar
from warnings import warn

from betty.concurrent import AsynchronizedLock, Lock
from betty.config import Configurable
from betty.typing import internal, public, Void, not_void, processsafe
from typing_extensions import override


@internal
class ServiceError(RuntimeError):
    """
    A service API error.
    """

    pass


@internal
class BootstrappedError(ServiceError):
    """
    Something was unexpectedly bootstrapped already.
    """

    pass


@internal
class NotBootstrappedError(ServiceError):
    """
    Something was unexpectedly not yet bootstrapped.
    """

    pass


@internal
class ServiceInitializedError(ServiceError):
    """
    A service was unexpectedly initialized already.
    """

    pass


@internal
class Bootstrapped:
    """
    A component that can be in a bootstrapped state.
    """

    def __init__(self, *args: Any, **kwargs: Any):
        self._bootstrapped = False
        super().__init__(*args, **kwargs)

    @final
    @property
    def bootstrapped(self) -> bool:
        """
        Whether the component has been bootstrapped.
        """
        return self._bootstrapped

    @final
    def assert_bootstrapped(self) -> None:
        """
        Assert that the component has been bootstrapped.
        """
        if not self.bootstrapped:
            raise NotBootstrappedError(f"{self} was not bootstrapped yet.")

    @final
    def assert_not_bootstrapped(self) -> None:
        """
        Assert that the component was not bootstrapped.
        """
        if self.bootstrapped:
            raise BootstrappedError(f"{self} was bootstrapped already.")


class Shutdownable(ABC):
    """
    A component that can be shut down.
    """

    @abstractmethod
    async def shutdown(self, *, wait: bool = True) -> None:
        """
        Shut the component down.
        """
        pass


class ShutdownCallbackKwargs(TypedDict):
    """
    The keyword arguments to a shutdown callback.
    """

    #: ``True`` to wait for the component to shut down gracefully, or ``False`` to attempt an immediate forced shutdown.
    wait: bool


ShutdownCallback: TypeAlias = Callable[
    [Unpack[ShutdownCallbackKwargs]], Awaitable[None]
]


@internal
@final
class ShutdownStack(Bootstrapped, Shutdownable):
    """
    A stack that invokes callbacks in reverse order upon shutting down.
    """

    def __init__(self):
        super().__init__()
        self._bootstrapped = True
        self._callbacks: MutableSequence[ShutdownCallback] = []

    @override
    async def shutdown(self, *, wait: bool = True) -> None:
        self.assert_bootstrapped()
        self._bootstrapped = False
        for callback in reversed(self._callbacks):
            await callback(wait=wait)

    def append(self, callback: ShutdownCallback | Shutdownable) -> None:
        """
        Append a callback or another component to the stack.
        """
        self._callbacks.append(
            callback.shutdown if isinstance(callback, Shutdownable) else callback
        )


@internal
class ServiceProvider(Bootstrapped, Shutdownable):
    """
    A service provider.

    Service providers make up a running Betty 'application'. They can provide services through
    :py:func:`betty.service.service`, and manage their resources by being bootstrapped and shut down.

    Service providers may be pickled once bootstrapped. Unpickled service providers are bootstrapped, and must be shut
    down by the caller.
    """

    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self._shutdown_stack = ShutdownStack()

    def __getstate__(self) -> dict[str, Any]:
        self.assert_bootstrapped()
        return {
            "_bootstrapped": True,
        }

    def __setstate__(self, state: dict[str, Any]) -> None:
        self.__dict__.update(state)
        self._shutdown_stack = ShutdownStack()

    @public
    async def bootstrap(self) -> None:
        """
        Bootstrap the component.
        """
        self.assert_not_bootstrapped()
        self._bootstrapped = True
        await self._bootstrap()

    async def _bootstrap(self) -> None:
        if isinstance(self, Configurable):
            self.configuration.immutable()
        await self._initialize_shared_services()

    async def _initialize_shared_services(self) -> None:
        """
        Initialize shared services, so they are ready to be pickled if/when they need to be.

        This is a workaround, because all pickling APIs are synchronous and will not allow us to call asynchronous
        service factories.
        """
        for _service_name, service_manager in getmembers(type(self)):
            if (
                isinstance(service_manager, ServiceManager)
                and service_manager.is_shared
            ):
                service = service_manager.get(self)
                if isinstance(service_manager, _AsynchronousServiceManager):
                    await service

    @public
    @override
    async def shutdown(self, *, wait: bool = True) -> None:
        self.assert_bootstrapped()
        self._bootstrapped = False
        await self._shutdown(wait=wait)

    async def _shutdown(self, *, wait: bool = True) -> None:
        await self._shutdown_stack.shutdown(wait=wait)
        if isinstance(self, Configurable):
            self.configuration.mutable()

    def __del__(self) -> None:
        if self.bootstrapped:
            warn(f"{self} was bootstrapped, but never shut down.", stacklevel=2)

    @public
    @final
    async def __aenter__(self) -> Self:
        await self.bootstrap()
        return self

    @public
    @final
    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        await self.shutdown(wait=exc_val is None)


_ServiceProviderT = TypeVar("_ServiceProviderT", bound=ServiceProvider)
_ServiceT = TypeVar("_ServiceT")
_ServiceGetT = TypeVar("_ServiceGetT")

ServiceFactory: TypeAlias = Callable[[_ServiceProviderT], _ServiceT]


@internal
class ServiceManager(Generic[_ServiceProviderT, _ServiceGetT, _ServiceT]):
    """
    Manages a single service for a service provider.
    """

    def __init__(
        self, factory: ServiceFactory[_ServiceProviderT, _ServiceGetT], *, shared: bool
    ):
        self._factory = factory
        self._shared = shared
        self._service_name = factory.__name__  # type: ignore[attr-defined]
        self._service_attr_name = f"_{self._service_name}"
        self._service_override_attr_name = f"{self._service_attr_name}_override"
        self._factory_override_attr_name = f"{self._service_attr_name}_factory_override"

    @property
    def is_shared(self) -> bool:
        """
        Whether the service is shared between service provider instances.
        """
        return self._shared

    @overload
    def __get__(self, instance: None, owner: type[_ServiceProviderT]) -> Self:
        pass

    @overload
    def __get__(
        self, instance: _ServiceProviderT, owner: type[_ServiceProviderT]
    ) -> _ServiceGetT:
        pass

    def __get__(
        self, instance: _ServiceProviderT | None, owner: type[_ServiceProviderT]
    ) -> _ServiceGetT | Self:
        if instance is None:
            return self  # type: ignore[return-value]

        return self.get(instance)

    def get(self, instance: _ServiceProviderT) -> _ServiceGetT:
        """
        Get the service from an instance.
        """
        instance.assert_bootstrapped()

        return self._get(instance)

    @abstractmethod
    def _get(self, instance: _ServiceProviderT) -> _ServiceGetT:
        pass

    def _get_attr(self, instance: _ServiceProviderT) -> _ServiceT | type[Void]:
        return getattr(instance, self._service_attr_name, Void)  # type: ignore[return-value]

    def _get_factory(
        self, instance: _ServiceProviderT
    ) -> ServiceFactory[_ServiceProviderT, _ServiceGetT]:
        factory = cast(
            "ServiceFactory[_ServiceProviderT, _ServiceGetT] | None",
            getattr(instance, self._factory_override_attr_name, None),
        )
        if factory is not None:
            return factory
        return self._factory

    def _assert_not_initialized(self, instance: _ServiceProviderT):
        if not_void(self._get_attr(instance)):
            raise ServiceInitializedError(
                f"{instance}.{self._service_name} was initialized already."
            )

    def override(self, instance: _ServiceProviderT, service: _ServiceT) -> None:
        """
        Override the service for the given instance.

        Calling this will prevent any existing factory from being called.

        This MUST only be called from ``instance.__init__()``.

        The provided service MUST be pickleable.
        """
        self._assert_not_initialized(instance)
        setattr(instance, self._service_attr_name, service)
        setattr(instance, self._service_override_attr_name, True)

    def override_factory(
        self,
        instance: _ServiceProviderT,
        factory: ServiceFactory[_ServiceProviderT, _ServiceGetT],
    ) -> None:
        """
        Override the default service factory for the given instance.

        This MUST only be called from ``instance.__init__()``. It will override the existing service factory method
        defined on the instance.

        The provided factory MUST be pickleable.
        """
        self._assert_not_initialized(instance)
        setattr(instance, self._factory_override_attr_name, factory)

    def get_state(self, instance: _ServiceProviderT) -> dict[str, Any]:
        """
        Get the attribute's state for the given instance.

        The returned state is the subset of ``instance.__dict__`` owned by this descriptor and that must be pickled
        along with ``instance``.
        """
        instance.assert_bootstrapped()
        service_overridden = getattr(instance, self._service_override_attr_name, False)
        if self.is_shared or service_overridden:
            service = self._get_attr(instance)
            assert not_void(service)
            return {
                self._service_attr_name: service,
            }
        factory = getattr(instance, self._factory_override_attr_name, None)
        if factory:
            return {
                self._factory_override_attr_name: factory,
            }
        return {}


class _AsynchronousServiceManager(
    Generic[_ServiceProviderT, _ServiceT],
    ServiceManager[_ServiceProviderT, Awaitable[_ServiceT], _ServiceT],
):
    def _lock(self, instance: _ServiceProviderT) -> Lock:
        lock_attr_name = f"_{self._service_attr_name}_lock"
        try:
            return cast("Lock", getattr(instance, lock_attr_name))
        except AttributeError:
            # We do not need a process-safe lock here, because we only ever lazily initialize services for the current
            # thread. Services that are shared across processes are explicitly initialized and pickled by service
            # providers.
            lock = AsynchronizedLock.threading()
            setattr(instance, lock_attr_name, lock)
            return lock

    async def _get(self, instance: _ServiceProviderT) -> _ServiceT:
        async with self._lock(instance):
            service = self._get_attr(instance)

            if not_void(service):
                return service

            new_service = await self._get_factory(instance)(instance)
            setattr(instance, self._service_attr_name, new_service)
            return new_service


class _SynchronousServiceManager(
    Generic[_ServiceProviderT, _ServiceT],
    ServiceManager[_ServiceProviderT, _ServiceT, _ServiceT],
):
    def _get(self, instance: _ServiceProviderT) -> _ServiceT:
        service = self._get_attr(instance)
        if not_void(service):
            return service

        new_service = self._get_factory(instance)(instance)
        setattr(instance, self._service_attr_name, new_service)
        return new_service


class _ServiceDecorator(Protocol):
    @overload
    def __call__(
        self, factory: Callable[[_ServiceProviderT], _ServiceT]
    ) -> _SynchronousServiceManager[_ServiceProviderT, _ServiceT]:
        pass

    @overload
    def __call__(
        self, factory: Callable[[_ServiceProviderT], Awaitable[_ServiceT]]
    ) -> _AsynchronousServiceManager[_ServiceProviderT, _ServiceT]:
        pass


@overload
def service(  # type: ignore[overload-overlap]
    factory: Callable[[_ServiceProviderT], Awaitable[_ServiceT]], shared: bool = False
) -> _AsynchronousServiceManager[_ServiceProviderT, _ServiceT]:
    pass


@overload
def service(
    factory: Callable[[_ServiceProviderT], _ServiceT], shared: bool = False
) -> _SynchronousServiceManager[_ServiceProviderT, _ServiceT]:
    pass


@overload
def service(factory: None = None, shared: bool = False) -> _ServiceDecorator:
    pass


def service(
    factory: Callable[[_ServiceProviderT], _ServiceGetT] | None = None,
    shared: bool = False,
) -> ServiceManager[_ServiceProviderT, _ServiceGetT, Any] | _ServiceDecorator:
    """
    Decorate a service factory method.

    The factory method is replaced with a :py:class:`service manager <betty.service.ServiceManager>` which handles lazy
    service instantiation, caching, and multiprocessing support.

    The decorated factory method should return a new service instance.
    """

    def _service(
        factory: Callable[[_ServiceProviderT], _ServiceGetT],
    ) -> ServiceManager[_ServiceProviderT, _ServiceGetT, Any]:
        if iscoroutinefunction(factory):
            return _AsynchronousServiceManager(factory, shared=shared)  # type: ignore[return-value]
        else:
            return _SynchronousServiceManager(factory, shared=shared)

    if factory is None:
        return _service  # type: ignore[return-value]
    return _service(factory)


@internal
@processsafe
class StaticService(Generic[_ServiceProviderT, _ServiceT]):
    """
    A service factory that returns a static, predefined service.
    """

    def __init__(self, service: _ServiceT):
        self._service = service

    def __call__(self, service_provider: _ServiceProviderT) -> _ServiceT:
        """
        Return the service.
        """
        return self._service
